!-------------------------------------------------------------------------------
!  AmplPicker.f90 - Created by Morgan Brown
!           Wed Feb 11 13:41:46 2004.
!-------------------------------------------------------------------------------
!       Usage: AmplPicker.x 
!  Parameters: 
! Description: 
!-------------------------------------------------------------------------------
program AmplPicker

use sep

implicit none

integer :: nt, nx1, nx2, ny1, ny2, ntwind, j0, j1, ix1, ix2, iy1, iy2, w0, iavo
real    :: dt, dx1, dx2, dy1, dy2, t0, x01, x02, y01, y02, b, d, e, f, g, ai, sinsq
logical :: outputavo

real, dimension(:)      , allocatable :: wd
real, dimension(:,:)    , allocatable :: tsurf
real, dimension(:,:,:)  , allocatable :: data, avo
real, dimension(:,:,:,:), allocatable :: maxampl


call sep_init(SOURCE)

call from_history('n1',nt)
call from_history('o1',t0)
call from_history('d1',dt)
call from_history('n2',nx1)
call from_history('d2',dx1)
call from_history('o2',x01)
call from_history('n3',nx2)
call from_history('d3',dx2)
call from_history('o3',x02)
call from_history('n4',ny1)
call from_history('d4',dy1)
call from_history('o4',y01)
call from_history('n5',ny2)
call from_history('d5',dy2)
call from_history('o5',y02)
call from_param('ntwind',ntwind,9)      ! Half width of analysis window
call from_param('w0',w0,0)              ! Shift origin of analysis window
call from_param('avo',iavo,0); outputavo=(iavo==1)

allocate( tsurf(ny1,ny2), data(nt,nx1,nx2), maxampl(nx1,nx2,ny1,ny2), avo(ny1,ny2,2), wd(nx1) )

call sep_read( tsurf, 'tsurf')

do iy2=1,ny2
	do iy1=1,ny1
		call sep_read( data)	
		j0 = min( nt, max(1,nint((tsurf(iy1,iy2)-t0)/dt)+1-ntwind+w0) )
		j1 = min( nt, max(1,nint((tsurf(iy1,iy2)-t0)/dt)+1+ntwind+w0) )
		write(0,"(a,2i4,a,2i4)") 'done ',iy1,iy2,' of ',ny1,ny2
		do ix2=1,nx2
			do ix1=1,nx1
				maxampl(ix1,ix2,iy1,iy2) = maxval( abs(data(j0:j1,ix1,ix2)) )
			end do
		end do

		wd = 1
		b=0; d=0; e=0; f=0; g=0
		do ix2=1,nx2
			do ix1=1,nx1
				ai = maxampl(ix1,ix2,iy1,iy2)
				if( ai==0.0) wd(ix1)=0.0
				sinsq = (ix1-1)*dx1 + x01
				b = b + sinsq*wd(ix1)
				d = d + sinsq*sinsq*wd(ix1)
				e = e + ai*sinsq
				f = f + ai
				g = g + wd(ix1)
			end do
		end do

		avo(iy1,iy2,1) = g*e - b*f
		avo(iy1,iy2,2) = -b*e + d*f
		if( abs(nx1*d - b*b) > 1e-6 ) then
			avo(iy1,iy2,:) = avo(iy1,iy2,:)/(nx1*d - b*b)
		else
			avo(iy1,iy2,:) = 0.0
		end if

	end do
end do

if( outputavo ) then
	call to_history('n1',ny1)
	call to_history('d1',dy1)
	call to_history('o1',y01)
	call to_history('n2',ny2)
	call to_history('d2',dy2)
	call to_history('o2',y02)
	call to_history('n3',2)
	call to_history('d3',1.0)
	call to_history('o3',0.0)
	call to_history('n4',1)
	call to_history('n5',1)
	call sep_write( avo )
else
	call to_history('n1',nx1)
	call to_history('d1',dx1)
	call to_history('o1',x01)
	call to_history('n2',nx2)
	call to_history('d2',dx2)
	call to_history('o2',x02)
	call to_history('n3',ny1)
	call to_history('d3',dy1)
	call to_history('o3',y01)
	call to_history('n4',ny2)
	call to_history('d4',dy2)
	call to_history('o4',y02)
	call to_history('n5',1)
	call sep_write( maxampl )
end if


call exit(0)

end program AmplPicker
