!-------------------------------------------------------------------------------
program CrosstalkModel3D_MPI

use sep
use mult_nmo_core_3d
use mult_nmo_util
use mpi_sep
use mult_nmo_mpi
use chain0_mod
use cartesian

implicit none

!-------------------------------------------------------------------------------
integer                         :: nt, nx1, nx2, ny1, ny2, ix1, ix2, stat,     &
                                   maxorder, nmod, iorder, itest, iverb,       &
                                   maxmem, nblock, nbout, jd0, jd1, jm0, jm1,  &
                                   iblock, ithr, nthr, iouter, nouter, iinner, &
                                   nbtmp, ninner, ilen, i0, i1, itauw, isplit, &
                                   korder, ksplit, ntot, it, jcmp(2), inorm,   &
                                   ncmp, ncmps, jv, maxoff1, maxoff2, nsurf,   &
                                   ksurf, isurf, ko0, ks0, j0, j1, jx1, jx2,   &
                                   ditauw, kcube, is1, isnell, jt, inaz, icmp, &
                                   inl
real                            :: dt, dx1, dx2, dy1, dy2, t0, x01, x02, y01,  &
                                   y02, tauw, refl, cutoff, r
logical                         :: testing, verb, norm, snell, narrow, nonlin
integer, dimension(:), pointer  :: isend, ilocal
real, dimension(:), pointer     :: vel, dip1, dip2, mblock, dblock, tmp,       &
                                   noffwt, dtmp, tauw_vect, refl_vect,         &
                                   noizmod_out, mtmp, x02_vect, vcorr
real, dimension(:,:,:), pointer :: tmp2d
character(len=256)              :: src, machine
character(len=20)               :: tmpstr
character(len=200)              :: model_tag, data_tag, vel_tag, dip1_tag,     &
                                   dip2_tag, refl_vect_tag, tauw_vect_tag,     &
                                   x02_vect_tag, vcorr_tag
!-------------------------------------------------------------------------------

#ifdef SEP_MPI 
	src = SOURCE
	if( .not. mpi_sep_setup(src,ithr,nthr,verb) ) call seperr("MPI_SEP_SETUP(): Error")
	call MPI_get_processor_name(machine,ilen,stat)	
	write(0,"(a,i2,a,a)") 'SEP_MPI IS DEFINED, ithr=',ithr,', Machine=',trim(machine)
#else
	write(0,*) 'SEP_MPI NOT DEFINED, ithr=',ithr
	ithr=0;  nthr=1
#endif

call from_param( "tauw"    , tauw    , 0.0)
call from_param( "refl"    , refl    , -1.0)
call from_param( "maxorder", maxorder )
call from_param( "cutoff"  , cutoff  , 0.05)
call from_param( "nsurf"   , nsurf   , 1)
call from_param( "maxmem"  , maxmem  , 200)   ! in MB
call from_param( "testing" , itest   , 0); testing=(itest==1)
call from_param( "norm"    , inorm   , 1); norm=(inorm==1)
call from_param( "snell"   , isnell  , 1); snell=(isnell==1)
call from_param( "narrow"  , inaz    , 0); narrow=(inaz==1)
call from_param( "nonlin"  , inl     , 0); nonlin=(inl==1)

call mult_nmo_mpi_init(verb)
nmod = get_nmod(maxorder,nsurf)

call fromaux_batch( "data", nt, dt, t0, nx1, dx1, x01, n3=nx2, d3=dx2, o3=x02,  &
                    n4=ny1, d4=dy1, o4=y01, n5=ny2, d5=dy2, o5=y02              )
if(ithr==0) call tohistory_batch( "model", nt, dt, t0, nx1, dx1, x01, n3=nx2,   &
                                  d3=dx2, o3=x02, n4=nmod, n5=ny1, d5=dy1,      &
                                  o5=y01, n6=ny2, d6=dy2, o6=y02                )
maxoff1 = min( nx1, 10)
maxoff2 = min( nx2, 10)
ncmp = nt*nx1*nx2
ncmps = ny1*ny2

nblock = nint( (maxmem*1000000)/(1.0*ncmp*(nmod+1)) )
nblock = min(nblock,ncmps)

allocate( mblock(ncmp*nmod*nblock), dblock(ncmp*nblock), tmp(ncmp),            &
          noffwt(ncmp*nmod), dtmp(ncmp), tmp2d(nt,nx1,nx2), mtmp(ncmp*nmod)    )

!-------------------------------------------------------------------------------
! Get filetags for this thread, i.e., "model_tmp_1" and "dip_tmp_3"
!-------------------------------------------------------------------------------
tmpstr="model";      call mpi_distrib_string( tmpstr, ithr, model_tag     )
tmpstr="data";       call mpi_distrib_string( tmpstr, ithr, data_tag      )
tmpstr="vel";        call mpi_distrib_string( tmpstr, ithr, vel_tag       )
tmpstr="x02_vect";   call mpi_distrib_string( tmpstr, ithr, x02_vect_tag  )
tmpstr="tauw_vect";  call mpi_distrib_string( tmpstr, ithr, tauw_vect_tag )
tmpstr="refl_vect";  call mpi_distrib_string( tmpstr, ithr, refl_vect_tag )
tmpstr="dip1";       call mpi_distrib_string( tmpstr, ithr, dip1_tag      )
tmpstr="dip2";       call mpi_distrib_string( tmpstr, ithr, dip2_tag      )
tmpstr="vcorr";      call mpi_distrib_string( tmpstr, ithr, vcorr_tag     )
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! Read and distribute auxiliary files
!-------------------------------------------------------------------------------
allocate( vel(nt*ncmps), tauw_vect(ncmps*nsurf), refl_vect(ncmps*nsurf),       &
          dip1(nt*ncmps), dip2(nt*ncmps), x02_vect(ncmps), vcorr(ncmps*nsurf)  )

call read_dist_aux( "vel", vel_tag, vel, ithr, nt*ncmps, test=testing,         &
                    check1=nt, check2=ny1, check3=ny2                          )
call read_dist_aux( "tauw_vect", tauw_vect_tag, tauw_vect, ithr, ncmps*nsurf,  &
                    test=testing, check1=ny1, check2=ny2, check3=nsurf,        &
                    default=tauw                                               )
call read_dist_aux( "refl_vect", refl_vect_tag, refl_vect, ithr, ncmps*nsurf,  &
                    test=testing, check1=ny1, check2=ny2, check3=nsurf,        &
                    default=refl                                               )
call read_dist_aux( "dip1", dip1_tag, dip1, ithr, nt*ncmps, test=testing,      &
                    check1=nt, check2=ny1, check3=ny2, default=0.0             )
call read_dist_aux( "dip2", dip2_tag, dip2, ithr, nt*ncmps, test=testing,      &
                    check1=nt, check2=ny1, check3=ny2, default=0.0             )
call read_dist_aux( "vcorr", vcorr_tag, vcorr, ithr, ncmps*nsurf,              &
                    test=testing, check1=ny1, check2=ny2, check3=nsurf,        &
                    default=1.0                                                )
if( narrow) then
	call read_dist_aux( "x02_vect", x02_vect_tag, x02_vect, ithr, ncmps,         &
	                    test=testing, check1=ny1, check2=ny2, default=x02        )
end if
!-------------------------------------------------------------------------------

call mult_nmo_core_3d_init( vel, tauw_vect, nt, t0, dt, nx1, x01, dx1, nx2,    &
                            x02, dx2, ny1, y01, dy1, ny2, y02, dy2, maxorder,  &
                            nsurf, refl_vect, dip1, dip2, vcorr, verb=verb,    &
                            testing=testing, ithread=ithr, narrow=narrow       )
if( nblock==ncmps ) then
	nouter = 1
else
	nouter = ncmps/nblock+1      ! integer math
end if

allocate( isend(nblock) )
if( .not. mpi_simple_distrib_pattern(isend,ilocal) ) call seperr("DIE DISTRIB")

do iouter=1,nouter
	nbtmp = min(nblock,ncmps-(iouter-1)*nblock)
	ninner=count(isend(1:nbtmp)==ithr)

	if( ninner>0 ) then
		write(0,*) '----------------------------------------------------'
		write(0,"(a,i2,a,i2,a,i2,a,i3,a,i3)") ' THREAD',ithr,': BLOCK ',&
		  iouter,' OF ',nouter,': BLOCKSZ=',nbtmp,'| NELEM in block=',ninner
		write(0,*) '----------------------------------------------------'
	
		call read_dist_main( "data", data_tag, dblock, ithr, ncmp, nbtmp, ninner,  &
		                     isend, testing                                        )
		do iblock=1,ninner
			iinner = ilocal(iblock)

			! Get cmpx/cmpy of cumulative CMP index; init modules
			call line2cart( (/ny1,ny2/), (iouter-1)*nblock+iinner, jcmp)
			icmp = grid2line2( (/ny1,ny2/), (/jcmp(1),jcmp(2)/) )
			if( narrow) call mnc_set_x02( x02_vect(icmp) )            ! Narrow azimuth

			call set_icmp( jcmp(1), jcmp(2))

			write(0,*) '  -----------------------------------------------------------'
			write(0,"(a,i2,a,i5,a,i5,a,i4)") '    thr=',ithr,' | CMPx=',jcmp(1),     &
			                                 ',CMPy=',jcmp(2),' | iblock=',iblock
			write(0,*) '  -----------------------------------------------------------'

			jd0 = grid2line2( (/ncmp,nblock/), (/1,   iblock/) )
			jd1 = grid2line2( (/ncmp,nblock/), (/ncmp,iblock/) )
			jm0 = grid2line2( (/ncmp*nmod,nblock/), (/1,        iblock/) )
			jm1 = grid2line2( (/ncmp*nmod,nblock/), (/ncmp*nmod,iblock/) )

			mblock(jm0:jm1) = 0.0

!-------------------------------------------------------------------------------
! cout_0  = N_0^T*(N_101 + N_111 + N_102 + N_112)*c_0
! cout_11 = ( N_101^T*(N_102 + N_112) + N_111^T*(N_102 + N_112) )*c_0
!-------------------------------------------------------------------------------

			!-------------------------------------------------------------------------
			! Step 1: NMO' the CMP gather.  Mute below WB tau*.  
			!-------------------------------------------------------------------------
			call set_order(0,0,1)
			stat = mult_nmo_exec( .true., .false., dtmp, dblock(jd0:jd1) )

			!-------------------------------------------------------------------------
			! Step 2: Loop over multiple legs and surfs.  Mute down to isurf tau*.
			!         Forward NMO to generate multiple model.  Populate mtmp this way.
			!-------------------------------------------------------------------------
			do iorder=1,maxorder
				do isplit=0,iorder
					do isurf=1,nsurf
						jv = grid2line3( (/ny1,ny2,nsurf/), (/jcmp(1),jcmp(2),1/) )
						itauw = nint( (tauw_vect(jv)-t0)/dt )+1                    ! WB tau*
						ditauw = nint( tauw_vect(jv)/dt ) +1
						tmp2d = reshape( dtmp, (/nt,nx1,nx2/) )
						jt = min( nt, max(1,itauw+ditauw+5) )
						if( .NOT. nonlin)  tmp2d(jt:nt,:,:) = 0.0     ! Zero below 2*WB tau*
						jv = grid2line3( (/ny1,ny2,nsurf/), (/jcmp(1),jcmp(2),isurf/) )
						r = refl_vect(jv)
						itauw = nint( (tauw_vect(jv)-t0)/dt )+1                 ! isurf tau*
						jt = min( nt, max(1,itauw-5) )
						tmp2d(1:jt,:,:) = 0.0                     ! Zero above isurf tau*
						kcube = get_cubeloc( iorder, nsurf, isurf, isplit)
						i0 = grid2line2( (/ncmp,nmod/), (/1,   kcube/) )
						i1 = grid2line2( (/ncmp,nmod/), (/ncmp,kcube/) )
						tmp = reshape( tmp2d, (/nt*nx1*nx2/) )
						call set_order(iorder,isplit,isurf)
						stat = mult_nmo_exec( .false., .false., tmp, mtmp(i0:i1) )
						mtmp(i0:i1) = mtmp(i0:i1)*abs(r)
					end do
				end do
			end do

			!-------------------------------------------------------------------------
			! Step 3: Loop over output crosstalk panels.  Add all crosstalk 
			!         contributions from all other multiple legs & surfs.
			!-------------------------------------------------------------------------
			do iorder=0,maxorder        ! Loop over output panels
				do isplit=0,iorder
					if(iorder==0) then
						is1=1
					else
						is1=nsurf
					end if

					if( nsurf>1 ) then
						ko0 = iorder
					else
						ko0 = iorder+1
					end if

					do isurf=1,is1

						call set_order(iorder,isplit,isurf)

						kcube = get_cubeloc( iorder, nsurf, isurf, isplit)
						j0 = grid2line2( (/ncmp,nmod/), (/1,   kcube/) )
						j1 = grid2line2( (/ncmp,nmod/), (/ncmp,kcube/) )

 						do korder=ko0,maxorder
							do ksplit=0,korder
								do ksurf=1,nsurf
									if( (iorder<maxorder) .OR. (isurf /= ksurf .AND. iorder>0) ) then
!write(0,"(a,i2,a,3i1,a,3i1)") 'THR',ithr,' adding ',korder,ksplit,ksurf,' to output ',iorder,isplit,isurf
										kcube = get_cubeloc( korder, nsurf, ksurf, ksplit)
										i0 = grid2line2( (/ncmp,nmod/), (/1,   kcube/) )
										i1 = grid2line2( (/ncmp,nmod/), (/ncmp,kcube/) )
										if( snell ) then
											call chain0( mult_nmo_exec, snell_resamp_exec, .true.,   &
											             .true., noffwt(j0:j1), mtmp(i0:i1), tmp     )
										else
											stat = mult_nmo_exec( .true., .true., noffwt(j0:j1),     &
											                      mtmp(i0:i1)                        )
										end if
									end if
								end do
							end do
						end do


!						where( noffwt(j0:j1) < cutoff*maxval(abs(noffwt(j0:j1))) ) noffwt(j0:j1) = 0
						if( norm .AND. maxval(abs(noffwt(j0:j1)))/=0.0 )                  &
							noffwt(j0:j1) = abs(noffwt(j0:j1))/maxval(abs(noffwt(j0:j1)))

					end do
				end do
			end do

			mblock(jm0:jm1) = noffwt

		end do

		call collect_write_main( "model", model_tag, mblock, ithr, nt,             &
		                         ncmp*nmod/nt, nbtmp, ninner, isend, test=testing  )
	end if

end do
!-------------------------------------------------------------------------------

call mult_nmo_core_close()

deallocate( vel, tauw_vect, refl_vect, dip1, dip2, mblock, dblock, tmp,        &
            noffwt, dtmp, tmp2d                           )

call mpi_finalize(stat)

call exit(0)

end program CrosstalkModel3D_MPI
