!-------------------------------------------------------------------------------
program DipSurfRegular

use sep
use polyfit

implicit none

integer :: nd, ny1, ny2, iy1, iy2, ndim, nsurf, nt
real    :: dy1, dy2, y01, y02, y1, y2, dt, t0

logical, parameter :: T=.true., F=.false.

real, dimension(:),       allocatable :: tvals, pvals
real, dimension(:,:),     allocatable :: txy
real, dimension(:,:,:),   allocatable :: tsurf
real, dimension(:,:,:,:), allocatable :: dipsurf, dipcube

call sep_init(SOURCE)

call from_history( "n1", ny1)
call from_history( "n2", ny2)
call from_history( "d1", dy1)
call from_history( "d2", dy2)
call from_history( "o1", y01)
call from_history( "o2", y02)
call from_history('n3',nsurf)
call from_history('n4',ndim)
if( ndim>1 .AND. ny2==1 ) call seperr('cannot have y-dip if n4>1')

call from_param( "nt",  nt)
call from_param( "dt",  dt)
call from_param( "t0",  t0)
call to_history( "n1", nt)
call to_history( "n2", ny1)
call to_history( "n3", ny2)
call to_history( "d1", dt)
call to_history( "d2", dy1)
call to_history( "d3", dy2)
call to_history( "o1", t0)
call to_history( "o2", y01)
call to_history( "o3", y02)
call to_history( "n4", 2)

allocate( txy(nd,3), tsurf(ny1,ny2,nsurf), dipsurf(ny1,ny2,nsurf,2),           &
          dipcube(nt,ny1,ny2,2)                                                )

write(0,*) 'bytes allocated=',nd*3+ny1*ny2*nsurf*3+nt*ny1*ny2*2

call sep_read( tsurf)
call sep_read( dipsurf)

!-------------------------------------------------------------------------------
! Interpolate regularly-sampled 3-D dip surfaces in time to output regularly-
! sampled dip cube
!-------------------------------------------------------------------------------
allocate( tvals(nsurf+2), pvals(nsurf+2) )
tvals(nsurf+2) = (nt-1)*dt + t0                       ! lower boundary condition
pvals(nsurf+2) = 0.0
tvals(1) = t0                                         ! upper boundary condition
pvals(1) = 0.0
do iy2=1,ny2
	do iy1=1,ny1
		tvals(2:nsurf+1) = tsurf(iy1,iy2,:)
		pvals(2:nsurf+1) = dipsurf(iy1,iy2,:,1)	
		call cubic_spline( tvals, pvals, nt, dt, t0, dipcube(:,iy1,iy2,1), F )
		tvals(2:nsurf+1) = tsurf(iy1,iy2,:)
		pvals(2:nsurf+1) = dipsurf(iy1,iy2,:,2)	
		call cubic_spline( tvals, pvals, nt, dt, t0, dipcube(:,iy1,iy2,2), F )
	end do
end do

call sep_write( dipcube)

deallocate( txy, tsurf, dipsurf, dipcube, tvals, pvals )

call exit(0)

end program DipSurfRegular
!-------------------------------------------------------------------------------
