! ---------------------------------------------------------------------------
!  DivideFiles.f90 - Created by Morgan Brown
!           Fri May 18 20:22:33 2001.
! ---------------------------------------------------------------------------
!       Usage: < data.H  DivideFiles.x itself[=1] [file2] 
!                                      zoz_replace[=0] > output.H
!  Parameters: itself[1]: Divide file by itself?    1=yes 0=no
!              file2: If defined, divide data.H by file2
!              zoz_replace: "Zero-over-zero replace" value.  Default: 0/0=0
! Description: Divide one file by itself or one file by another.  Set division
!              by zero to zoz_replace.  Reads in cubes at a time (n1:n3),
!              loops over outer planes (n4:n5).
! ---------------------------------------------------------------------------
program DivideFiles

use sep

implicit none

integer                    :: itself, iouter, ndim, nouter, ninner
real                       :: d1, d2, o1, o2, zoz_replace
integer, dimension(:), allocatable :: n
real, dimension(:), allocatable :: data, data2, output 

call sep_init(SOURCE)

ndim = sep_dimension()
allocate( n(ndim) )

call from_history( 'n', n )
call from_param( 'itself', itself, 1)
call from_param( 'zoz_replace', zoz_replace, 0.0)

nouter = product( n(min(4,ndim):ndim) )
ninner = product(n)/nouter

allocate( data(ninner), output(ninner) )

if( itself==0 ) allocate( data2(ninner) )

do iouter=1,nouter
	write(0,"(a,i4,a,i4)") 'done outer block',iouter,' of',nouter
	data=0.0;  call sep_read( data)

	output=zoz_replace
	if( itself==0 ) then
		data2 = 0.0
		call sep_read( data2, "file2" )
		where(data2/=0) output = data/data2
	else
		where(data/=0) output = data/data
	end if

	call sep_write( output)
end do

deallocate( data, output )
if( itself==0 ) deallocate( data2 )

call exit(0)

end program DivideFiles
