program InterpSurf
!--------------------------------------------------------------------------------
! Program InterpRMS
!--------------------------------------------------------------------------------
! Usage:  InterpRMS.x o1= d1= n1= < pickfile.H > interp.H
!--------------------------------------------------------------------------------
! Description:  
!--------------------------------------------------------------------------------
! Parameters:  n1 - Size of output model grid.
!              o1 - Origin of output model grid.
!              d1 - Grid spacing of output model grid.
!--------------------------------------------------------------------------------
! Input: (pickfile.H) A file of time coordinates/velocity pick pairs. n2=3
!--------------------------------------------------------------------------------
! Output: (interp.H) RMS velocity on a regular grid defined by o1, d1, n1.
!--------------------------------------------------------------------------------
  use sep
	use solver_reg_mod
	use cgstep_mod
	use igrad2
	use bin2
	use grad2fill

  implicit none
  integer                        :: nd, m1, m2, niter, stat, niter
  integer                        :: iverb, ireg
  real, dimension (:),   pointer :: mm, t
  real, dimension (:,:), pointer :: xy
  real                           :: o1, d1, o2, d2, eps
	logical                        :: verb, reg

  call sep_init(SOURCE)

  call from_param( "n1", m1);  call from_param( "o1", o1, 0.0);  call from_param( "d1", d1, 1.0)
  call from_param( "n2", m2);  call from_param( "o2", o2, 0.0);  call from_param( "d2", d2, 1.0)
	call from_param( "niter", niter, m1*m2);
	call from_param( "eps", eps, 1.0);
	call from_param( "verb", iverb, 0);  verb=(iverb==1)
	call from_param( "reg",  ireg,  0);  reg=(ireg==1)

  call from_history("n1", nd)
  call to_history( "n1", m1);  call to_history( "o1", o1);  call to_history( "d1", d1)
  call to_history( "n2", m2);  call to_history( "o2", o2);  call to_history( "d2", d2)

	allocate ( mm(m1*m2), t(nd), xy(nd,2) )

	call sep_read( t);  call sep_read( xy(:,1));  call sep_read( xy(:,2))

	call bin2_init( m1, m2, o1, d1, o2, d2, xy)
	call igrad2_init( m1, m2)
	stat = bin2_lop( .true., .false., mm, t)

	if( reg) then
		call solver_reg( mm, t, bin2_lop, igrad2_lop, cgstep, 2*m1*m2, niter, eps, verb=verb, m0=mm)
	else
		call grad2fill2( niter, m1, m2, mm, (mm>0) )
	end if

	call sep_write(mm)

  deallocate( mm, t, xy)

  call exit(0) 

end program InterpSurf
