!-------------------------------------------------------------------------------
! Losmooth - Local-window smoothing in up to 3-D.  
!            method=1: median 
!            method=2: weighted mean
!-------------------------------------------------------------------------------
program Losmooth

	use sep
	use quantile_mod

	implicit none
	
	integer :: n123, nw123, n4, n5, stat, ipch, iverb, method, i1, i2, i3, i4, i5, c
	logical :: verb
	integer :: n(3), nwin(3), mwin(3), i(3), j(3), k(3), kk(3)
	real, pointer, dimension(:)     :: tmp
	real, pointer, dimension(:,:,:) :: d, w, m, dwin, wwin

	call sep_init(SOURCE)

	call from_history( "n1", n(1) )
	call from_history( "n2", n(2) )
	call from_history( "n3", n(3) )
	call from_history( "n4", n4   )
	call from_history( "n5", n5   )
	call from_param( "nwin1", nwin(1), min(4,n(1)) )
	call from_param( "nwin2", nwin(2), min(4,n(2)) )
	call from_param( "nwin3", nwin(3), min(4,n(3)) )
	call from_param( "method", method, 1 )
	call from_param( "verb" , iverb  , 0 );  verb = (iverb==1)

	if(verb) write(0,*) 'Losmooth.x - top'

	n123  = product( n )
	
	if(verb) write(0,*) 'data size:',n
	if(verb) write(0,*) 'window size:',nwin

	mwin = nwin*2 + 1

	allocate( d(n(1),n(2),n(3)), w(n(1),n(2),n(3)), m(n(1),n(2),n(3)) )
	allocate( dwin(mwin(1),mwin(2),mwin(3)), wwin(mwin(1),mwin(2),mwin(3)) )

	do i5=1,n5
		do i4=1,n4

			if(verb) write(0,"(a,i3,a,i3,a,i3,a,i3)") "outer block",i4,",",i5," of ",n4,",",n5

			if( exist_file("wght") ) then
				call sep_read( w, "wght" )	
			else
				w = 1.0
			end if

			call sep_read( d )

			do i3=1,n(3)
				do i2=1,n(2)
					do i1=1,n(1)

						i = (/i1,i2,i3/)
						j = max( 1, i - nwin )
						k = min( n, i + nwin )
						kk = k - j + 1

						dwin = 0.0;  wwin = 0.0

						dwin(1:kk(1),1:kk(2),1:kk(3)) = d(j(1):k(1),j(2):k(2),j(3):k(3))
						wwin(1:kk(1),1:kk(2),1:kk(3)) = w(j(1):k(1),j(2):k(2),j(3):k(3))

						if( method==1) then                               ! Median Filter 
							nw123 = product( kk )
							allocate( tmp(nw123) )
							tmp = reshape( dwin, (/nw123/) )
							m(i1,i2,i3) = quantile( nw123/2, tmp )    
							deallocate( tmp )
						else                                              ! Wght. Mean
							if( sum(wwin)==0.0 ) wwin(1:kk(1),1:kk(2),1:kk(3))=1.0 
							m(i1,i2,i3) = sum(dwin*wwin)/sum(wwin) 
						end if

					end do
				end do
			end do
	
			call sep_write( m )

		end do
	end do

	deallocate( d, w, m, wwin, dwin)

	call exit(0)

end program Losmooth
