!-------------------------------------------------------------------------------
program MultIndivOper3D_MPI

use sep
use mult_nmo_core_3d
use mult_nmo_ampl_3d
use mult_nmo_util
use mpi_sep
use mult_nmo_mpi
use cartesian

implicit none

!-------------------------------------------------------------------------------
integer :: nt, nx1, nx2, ny1, ny2, stat, maxorder, ncmp, niter, ncmps, ichain, &
           chain(:), noper, igsc, irefl, itest, iverb, maxmem, nblock, nbout,  &
           j0, j1, iblock, jcmp(2), ithr, nthr, iouter, nouter, iinner, nbtmp, &
           ninner, ilen, order, split, nsurf, inaz, icmp, idip, isnorm
real    :: dt, dx1, dx2, dy1, dy2, t0, x01, x02, y01, y02, tauw, refl
logical :: testing, verb, invrefl, gscinv, narrow, dipping, snellnorm

logical, dimension(:,:), pointer :: misstrace
integer, dimension(:), pointer :: isend, ilocal
real, dimension(:), pointer    :: vel, reswt, dip1, dip2, mblock, dblock,      &
                                  tauw_vect, data, model, refl_vect, tmp,      &
                                  mones, dones, x02_vect, vcorr
character(len=256)             :: src, machine
character(len=20)              :: tmpstr
character(len=200)             :: model_tag, data_tag, vel_tag, tauw_vect_tag, &
                                  refl_vect_tag, dip1_tag, dip2_tag,           &
                                  x02_vect_tag, vcorr_tag
#ifdef SEP_MPI 
	src = SOURCE
	if( .not. mpi_sep_setup(src,ithr,nthr,verb) )                                &
	    call seperr("MPI_SEP_SETUP(): Error")
	call MPI_get_processor_name(machine,ilen,stat)	
	write(0,"(a,i2,a,a)") 'SEP_MPI IS DEFINED, ithr=',ithr,', Machine=',         &
	                      trim(machine)
#else
	write(0,*) 'SEP_MPI NOT DEFINED, ithr=',ithr
	ithr=0;  nthr=1
#endif

call from_param( "refl"     , refl  , -1.0)
call from_param( "order"    , order , 0)
call from_param( "split"    , split , 0)
call from_param( "nsurf"    , nsurf , 1)
call from_param( "noper"    , noper , 1)
call from_param( "maxmem"   , maxmem, 200)   ! in MB
call from_param( "testing"  , itest , 0); testing=(itest==1)
call from_param( "verb"     , iverb , 0); verb=(iverb==1)
call from_param( "invrefl"  , irefl , 0); invrefl=(irefl==1)
call from_param( "gscinv"   , igsc  , 0); gscinv=(igsc==1)
call from_param( "narrow"   , inaz  , 0); narrow=(inaz==1)
call from_param( "dipping"  , idip  , 1); dipping=(idip==1)
call from_param( "snellnorm", isnorm, 1); snellnorm=(isnorm==1)

allocate( chain(noper) );  chain=0
call from_param( "chain", chain)               ! chain=0,1 -> NMO + Snell

call mult_nmo_mpi_init(verb)

!-------------------------------------------------------------------------------
call fromaux_batch( "data", nt, dt, t0, nx1, dx1, x01, n3=nx2, d3=dx2, o3=x02, &
                     n4=ny1, d4=dy1, o4=y01, n5=ny2, d5=dy2, o5=y02            )

call from_param( "tauw", tauw, t0)         ! No need to specify tauw for order=0

if(ithr==0) call tohistory_batch( "model", nt, dt, t0, nx1, dx1, x01, n3=nx2,  &
                                  d3=dx2, o3=x02, n4=ny1, d4=dy1, o4=y01,      &
                                  n5=ny2, d5=dy2, o5=y02                       )
ncmp = nt*nx1*nx2                                              ! Size of one CMP
ncmps = ny1*ny2                                                ! Total # of CMPs
nblock = nint( (maxmem*1000000)/(2.0*ncmp) )
nblock = min(nblock,ny1*ny2)

allocate( mblock(ncmp*nblock), dblock(ncmp*nblock), mones(ncmp), dones(ncmp),  &
          tmp(ncmp), misstrace(nx1,nx2)                                        )

!-------------------------------------------------------------------------------
! Get filetags for this thread, i.e., "model_tmp_1" and "dip_tmp_3"
!-------------------------------------------------------------------------------
tmpstr="model";      call mpi_distrib_string( tmpstr, ithr, model_tag     )
tmpstr="data";       call mpi_distrib_string( tmpstr, ithr, data_tag      )
tmpstr="vel";        call mpi_distrib_string( tmpstr, ithr, vel_tag       )
tmpstr="x02_vect";   call mpi_distrib_string( tmpstr, ithr, x02_vect_tag  )
tmpstr="tauw_vect";  call mpi_distrib_string( tmpstr, ithr, tauw_vect_tag )
tmpstr="refl_vect";  call mpi_distrib_string( tmpstr, ithr, refl_vect_tag )
tmpstr="dip1";       call mpi_distrib_string( tmpstr, ithr, dip1_tag      )
tmpstr="dip2";       call mpi_distrib_string( tmpstr, ithr, dip2_tag      )
tmpstr="vcorr";      call mpi_distrib_string( tmpstr, ithr, vcorr_tag     )
! ------------------------------------------------------------------------------

! ------------------------------------------------------------------------------
! Read and distribute auxiliary files
! ------------------------------------------------------------------------------
allocate( vel(nt*ncmps), tauw_vect(ncmps*nsurf), refl_vect(ncmps*nsurf),       &
          dip1(nt*ncmps), dip2(nt*ncmps), x02_vect(ncmps), vcorr(ncmps*nsurf)  )

if(verb) write(0,*) 'bytes allocated=',2*ncmp*nblock+3*ncmp+3*nt*ncmps+2*ncmps*nsurf+ncmps

call read_dist_aux( "vel", vel_tag, vel, ithr, nt*ncmps, test=testing,         &
                    check1=nt, check2=ny1, check3=ny2                          )

call read_dist_aux( "tauw_vect", tauw_vect_tag, tauw_vect, ithr, ncmps*nsurf,  &
                    test=testing, check1=ny1, check2=ny2, check3=nsurf,        &
                    default=tauw                                               )
call read_dist_aux( "refl_vect", refl_vect_tag, refl_vect, ithr, ncmps*nsurf,  &
                    test=testing, check1=ny1, check2=ny2, check3=nsurf,        &
                    default=refl                                               )
call read_dist_aux( "dip1", dip1_tag, dip1, ithr, nt*ncmps, test=testing,      &
                    check1=nt, check2=ny1, check3=ny2, default=0.0             )
call read_dist_aux( "dip2", dip2_tag, dip2, ithr, nt*ncmps, test=testing,      &
                    check1=nt, check2=ny1, check3=ny2, default=0.0             )
call read_dist_aux( "vcorr", vcorr_tag, vcorr, ithr, ncmps, test=testing,      &
                    check1=ny1, check2=ny2, check3=nsurf, default=1.0          )
if( narrow) then
	call read_dist_aux( "x02_vect", x02_vect_tag, x02_vect, ithr, ncmps,         &
	                    test=testing, check1=ny1, check2=ny2, default=x02        )
end if
! ------------------------------------------------------------------------------

call mult_nmo_core_3d_init( vel, tauw_vect, nt, t0, dt, nx1, x01, dx1, nx2,    &
                            x02, dx2, ny1, y01, dy1, ny2, y02, dy2, order,     &
                            nsurf, refl_vect, dip1, dip2, vcorr, verb=verb,    &
                            testing=testing, invrefl=invrefl, ithread=ithr,    &
                            invgsc=gscinv, narrow=narrow, dipping=dipping,     &
                            normsnell=snellnorm                                )
call set_order(order,split,nsurf)

if( nblock==ncmps ) then
	nouter = 1
else
	nouter = ncmps/nblock+1      ! integer math
end if

allocate( isend(nblock) )
if( .not. mpi_simple_distrib_pattern(isend,ilocal) ) call seperr("DIE DISTRIB")

! ---------LOOP OVER OUTER BLOCKS-----------------------------------------------
do iouter=1,nouter
	nbtmp = min(nblock,ncmps-(iouter-1)*nblock)       ! #CMPS in this outer block
	ninner = count(isend(1:nbtmp)==ithr)              ! #CMPS owned by this thread

	if( ninner>0 ) then
		write(0,*) '---------------------------------------------------------------'
		write(0,"(a,i2,a,i2,a,i2,a,i3,a,i3)") ' THREAD',ithr,': BLOCK ',&
		  iouter,' OF ',nouter,': BLOCKSZ=',nbtmp,'| NELEM in block=',ninner
		write(0,*) '---------------------------------------------------------------'
	
		call read_dist_main( "data", data_tag, dblock, ithr, ncmp, nbtmp, ninner,  & 
		                     isend, testing                                        )

		! ---------LOOP OVER INNER BLOCKS-------------------------------------------
		do iblock=1,ninner
			iinner = ilocal(iblock)                       ! Index in block of this CMP

			! Get cmpx/cmpy of cumulative CMP index; init modules
			call line2cart( (/ny1,ny2/), (iouter-1)*nblock+iinner, jcmp)
			icmp = grid2line2( (/ny1,ny2/), (/jcmp(1),jcmp(2)/) )
			if( narrow) call mnc_set_x02( x02_vect(icmp) )            ! Narrow azimuth

			write(0,*) '  -----------------------------------------------------------'
			write(0,"(a,i2,a,i5,a,i5,a,i4)") '    thr=',ithr,' | CMPx=',jcmp(1),     &
			                                 ',CMPy=',jcmp(2),' | iblock=',iblock
			write(0,*) '  -----------------------------------------------------------'

			j0 = grid2line2( (/ncmp,nblock/), (/1,   iblock/) )
			j1 = grid2line2( (/ncmp,nblock/), (/ncmp,iblock/) )

			call get_misstrace( nt, nx1, nx2, dblock(j0:j1), misstrace)
			call set_icmp( jcmp(1), jcmp(2), misstrace=misstrace)
	
			! ------MAIN BLOCK -------------------------------------------------------
			tmp = dblock(j0:j1)
			do ichain=1,noper
				if( chain(ichain)==0 ) then
					stat = mult_nmo_exec( .true., .false., mblock(j0:j1), tmp)
				else if( chain(ichain)==1 ) then
					stat = snell_resamp_exec( .true., .false., mblock(j0:j1), tmp)
				else if( chain(ichain)==2 ) then
					stat = geometric_spreading_exec( .true., .false., mblock(j0:j1), tmp)
				else if( chain(ichain)==3 ) then
					stat = onset_mute_exec( .true., .false., mblock(j0:j1), tmp)
				end if
				tmp = mblock(j0:j1)
			end do
			! ------------------------------------------------------------------------
	
		! -----END LOOP OVER INNER BLOCKS-------------------------------------------
		end do

		call collect_write_main( "model", model_tag, mblock, ithr, nt, ncmp/nt,    &
		                         nbtmp, ninner, isend, test=testing                )

	end if !----(ninner > 0)------------------------------------------------------
end do
! -----END LOOP OVER OUTER BLOCKS-----------------------------------------------

call mult_nmo_core_close()

deallocate( vel, tauw_vect, refl_vect, dip1, dip2, mblock, dblock, isend )

call mpi_finalize(stat)

call exit(0)

end program MultIndivOper3D_MPI
