!-------------------------------------------------------------------------------
program MultNmo3D_MPI

use sep
use solver_reg_mod
use cgstep_mod
use mult_nmo_core_3d
use mult_nmo_reg_3d
use mult_nmo_util
use weight
use mpi_sep
use mult_nmo_mpi
use cartesian

implicit none

!-------------------------------------------------------------------------------
integer                        :: nt, nx1, nx2, ny1, ny2, stat, maxorder,      &
                                  niter, ncmp, i3, ncmps, imod, ioptim, igsc,  &
                                  irefl, itest, iverb, nmod, ny1_vel, ny1_tmp, &
                                  ny2_vel, ny2_tmp, maxmem, nblock, nbout,     &
                                  nbytes, jm0, jm1, jd0, jd1, iblock, jcmp(2), &
                                  ithr, nthr, iouter, nouter, iinner, nbtmp,   &
                                  ninner, ilen, idipping, iymin1, iymax1,      &
                                  iymin2, iymax2, ndone, nsurf, isnell, im,    &
                                  im0, im1, ix, it, isnorm, inaz, icmp, nstk,  &
                                  cmpres(2)
real                           :: dt, dx1, dx2, dy1, dy2, t0, x01, x02, y01,   &
                                  y02, tauw, refl, eps, eps_order,         &
                                  eps_grad, eps_xtalk
integer, dimension(:), pointer :: isend, ilocal, nullcube
real, dimension(:), pointer    :: vel, reswt, dip1, dip2, mblock, dblock,      &
                                  xblock, tauw_vect, data, model, refl_vect,   &
                                  crosstalk, x02_vect, rd, rm, vcorr
real, dimension(:,:), pointer  :: rmov, mmov
logical                        :: optim, modeling, testing, verb1, verb2,      &
                                  invrefl,  dipping, snell, gsc, snellnorm,    &
                                  narrow
character(len=256)             :: src, machine
character(len=20)              :: tmpstr
character(len=200)             :: model_tag, data_tag, crosstalk_tag, vel_tag, &
                                  tauw_vect_tag, refl_vect_tag, dip1_tag,      &
                                  dip2_tag, x02_vect_tag, vcorr_tag, m0_tag

logical, parameter :: T = .true., F = .false.

#ifdef SEP_MPI 
	src = SOURCE
	if( .not. mpi_sep_setup(src,ithr,nthr,verb2) ) call seperr("MPI_SEP_SETUP(): Error")
	call MPI_get_processor_name(machine,ilen,stat)	
	write(0,"(a,i2,a,a)") 'SEP_MPI IS DEFINED, ithr=',ithr,', Machine=',trim(machine)
#else
	write(0,*) 'SEP_MPI NOT DEFINED, ithr=',ithr
	ithr=0;  nthr=1
#endif

call from_param( "tauw"     , tauw     , 0.0)
call from_param( "refl"     , refl     , -1.0)
call from_param( "maxorder" , maxorder , 0)
call from_param( "niter"    , niter    , nt/50)
call from_param( "nsurf"    , nsurf    , 1)
call from_param( "eps"      , eps      , 1.0)
call from_param( "eps_order", eps_order, 1.0)
call from_param( "eps_grad" , eps_grad , 1.0)
call from_param( "eps_xtalk", eps_xtalk, 1.0)
call from_param( "maxmem"   , maxmem   , 200)   ! in MB
call from_param( "optim"    , ioptim   , 0); optim=(ioptim==1)
call from_param( "modeling" , imod     , 0); modeling=(imod==1)
call from_param( "testing"  , itest    , 0); testing=(itest==1)
call from_param( "verb"     , iverb    , 0); verb1=(iverb>0); verb2=(iverb>1)
call from_param( "invrefl"  , irefl    , 0); invrefl=(irefl==1)
call from_param( "dipping"  , idipping , 1); dipping=(idipping==1)
call from_param( "snell"    , isnell   , 1); snell=(isnell==1)
call from_param( "gsc"      , igsc     , 1); gsc=(igsc==1)
call from_param( "snellnorm", isnorm   , 0); snellnorm=(isnorm==1)
call from_param( "narrow"   , inaz     , 0); narrow=(inaz==1)

call mult_nmo_mpi_init(verb2)
nmod=get_nmod(maxorder,nsurf)

if(testing .AND. ithr==0 ) then
	call mult_nmo_core_3d_test()
	call mult_nmo_reg_3d_test()
end if

!-------------------------------------------------------------------------------
! Read model or data, depending on whether we're running LSJIMP, adjoint, or
! forward model.
!-------------------------------------------------------------------------------
if(modeling .AND. (.NOT. optim) ) then
	call fromaux_batch( "model", nt, dt, t0, nx1, dx1, x01, n3=nx2, d3=dx2,      &
	                    o3=x02, n4=nmod, n5=ny1, d5=dy1, o5=y01, n6=ny2, d6=dy2, &
	                    o6=y02                                                   )
	if(ithr==0) call tohistory_batch( "data", nt, dt, t0, nx1, dx1, x01, n3=nx2, &
	                                  d3=dx2, o3=x02, n4=ny1, d4=dy1, o4=y01,    &
	                                  n5=ny2, d5=dy2, o5=y02                     )
else
	call fromaux_batch( "data", nt, dt, t0, nx1, dx1, x01, n3=nx2, d3=dx2,       &
	                    o3=x02, n4=ny1, d4=dy1, o4=y01, n5=ny2, d5=dy2, o5=y02   )
	if(ithr==0) call tohistory_batch( "model", nt, dt, t0, nx1, dx1, x01,        &
	                                  n3=nx2, d3=dx2, o3=x02, n4=nmod, n5=ny1,   &
	                                  d5=dy1, o5=y01, n6=ny2, d6=dy2, o6=y02     )
end if

!-------------------------------------------------------------------------------
! Option to process limited midpoint range
!-------------------------------------------------------------------------------
call from_param( "iymin1", iymin1, 1  )
call from_param( "iymin2", iymin2, 1  )
call from_param( "iymax1", iymax1, ny1)
call from_param( "iymax2", iymax2, ny2)
cmpres = (/1,1/)
call from_param( "cmpres", cmpres, cmpres)

ncmp = nt*nx1*nx2                                              ! Size of one CMP
ncmps = ny1*ny2                                                ! Total # of CMPs

!-------------------------------------------------------------------------------
! Set blocksize (in CMPs) depending on max memory allowed (maxmem)
!-------------------------------------------------------------------------------
if(optim) then
	nblock = nint( (maxmem*1000000)/(1.0*ncmp*(2*nmod+1)) )
else
	nblock = nint( (maxmem*1000000)/(1.0*ncmp*(nmod+1)) )
end if
nblock = min(nblock,ny1*ny2)

allocate( mblock(ncmp*nmod*nblock), dblock(ncmp*nblock), reswt(ncmp),          &
          rd(ncmp), rm(ncmp*nmod*4), model(ncmp*nmod), data(ncmp),             &
          rmov(ncmp*(1+nmod*4),niter), mmov(ncmp*nmod,niter)                   )
if(optim)  allocate( xblock(ncmp*nmod*nblock), nullcube(ncmp*nmod) )

!-------------------------------------------------------------------------------
! Get filetags for this thread, i.e., "model_tmp_1" and "dip_tmp_3"
!-------------------------------------------------------------------------------
tmpstr="model";      call mpi_distrib_string( tmpstr, ithr, model_tag     )
tmpstr="data";       call mpi_distrib_string( tmpstr, ithr, data_tag      )
tmpstr="crosstalk";  call mpi_distrib_string( tmpstr, ithr, crosstalk_tag )
tmpstr="m0";         call mpi_distrib_string( tmpstr, ithr, m0_tag        )
tmpstr="vel";        call mpi_distrib_string( tmpstr, ithr, vel_tag       )
tmpstr="x02_vect";   call mpi_distrib_string( tmpstr, ithr, x02_vect_tag  )
tmpstr="tauw_vect";  call mpi_distrib_string( tmpstr, ithr, tauw_vect_tag )
tmpstr="refl_vect";  call mpi_distrib_string( tmpstr, ithr, refl_vect_tag )
tmpstr="dip1";       call mpi_distrib_string( tmpstr, ithr, dip1_tag      )
tmpstr="dip2";       call mpi_distrib_string( tmpstr, ithr, dip2_tag      )
tmpstr="vcorr";      call mpi_distrib_string( tmpstr, ithr, vcorr_tag     )
! ------------------------------------------------------------------------------

! ------------------------------------------------------------------------------
! Read and distribute auxiliary files
! ------------------------------------------------------------------------------
allocate( vel(nt*ncmps), tauw_vect(ncmps*nsurf), refl_vect(ncmps*nsurf),       &
          dip1(nt*ncmps), dip2(nt*ncmps), x02_vect(ncmps), vcorr(ncmps*nsurf)  )

call read_dist_aux( "vel", vel_tag, vel, ithr, nt*ncmps, test=testing,         &
                    check1=nt, check2=ny1, check3=ny2                          )
call read_dist_aux( "tauw_vect", tauw_vect_tag, tauw_vect, ithr, nsurf*ncmps,  &
                    test=testing, check1=ny1, check2=ny2, check3=nsurf,        &
                    default=tauw                                               )
call read_dist_aux( "refl_vect", refl_vect_tag, refl_vect, ithr, nsurf*ncmps,  &
                    test=testing, check1=ny1, check2=ny2, check3=nsurf,        &
                    default=refl                                               )
call read_dist_aux( "dip1", dip1_tag, dip1, ithr, nt*ncmps, test=testing,      &
                    check1=nt, check2=ny1, check3=ny2, default=0.0             )
call read_dist_aux( "dip2", dip2_tag, dip2, ithr, nt*ncmps, test=testing,      &
                    check1=nt, check2=ny1, check3=ny2, default=0.0             )
call read_dist_aux( "vcorr", vcorr_tag, vcorr, ithr, ncmps*nsurf,              &
                    test=testing, check1=ny1, check2=ny2, check3=nsurf,        &
                    default=1.0                                                )
if( narrow) then
	call read_dist_aux( "x02_vect", x02_vect_tag, x02_vect, ithr, ncmps,         &
	                    test=testing, check1=ny1, check2=ny2, default=x02        )
end if
! ------------------------------------------------------------------------------

call mult_nmo_core_3d_init( vel, tauw_vect, nt, t0, dt, nx1, x01, dx1, nx2,    &
                            x02, dx2, ny1, y01, dy1, ny2, y02, dy2, maxorder,  &
                            nsurf, refl_vect, dip1, dip2, vcorr, verb=verb2,   &
                            testing=testing, invrefl=invrefl, ithread=ithr,    &
                            dipping=dipping, do_snell=snell, do_gsc=gsc,       &
                            normsnell=snellnorm, narrow=narrow                 )

if(optim) call mult_nmo_reg_init( nt, nx1, nx2, maxorder, nsurf, eps_order,    &
                                  eps_grad, eps_xtalk, verb=verb2, ithread=ithr)

! ------------------------------------------------------------------------------
! Compute number of blocks to process
! ------------------------------------------------------------------------------
if( nblock==ncmps ) then
	nouter = 1
else
	nouter = ncmps/nblock+1      ! integer math
end if

allocate( isend(nblock) )
if( .not. mpi_simple_distrib_pattern(isend,ilocal) ) call seperr("DIE DISTRIB")

ndone = 0

! ---------LOOP OVER OUTER BLOCKS-----------------------------------------------
do iouter=1,nouter
	nbtmp = min(nblock,ncmps-(iouter-1)*nblock)       ! #CMPS in this outer block
	ninner = count(isend(1:nbtmp)==ithr)              ! #CMPS owned by this thread

	if( ninner>0 ) then
		write(0,*) '---------------------------------------------------------------'
		write(0,"(a,i2,a,i2,a,i2,a,i3,a,i3)") ' THREAD',ithr,': BLOCK ',&
		  iouter,' OF ',nouter,': BLOCKSZ=',nbtmp,'| NELEM in block=',ninner
		write(0,*) '---------------------------------------------------------------'
	
		if( modeling .AND. (.NOT. optim) ) then
			call read_dist_main( "model", model_tag, mblock, ithr, ncmp*nmod, nbtmp, &
			                     ninner, isend, testing                              )
		else
			call read_dist_main( "data", data_tag, dblock, ithr, ncmp, nbtmp, ninner,& 
			                     isend, testing                                      )
			if( optim ) then
				if( exist_file("crosstalk") ) then
					call read_dist_main( "crosstalk", crosstalk_tag, xblock, ithr,       & 
					                     ncmp*nmod, nbtmp, ninner, isend, testing        )
				else
					xblock = 0.0
				end if

				if( exist_file("m0") ) then
					call read_dist_main( "m0", m0_tag, mblock, ithr, ncmp*nmod, nbtmp,   &
					                     ninner, isend, testing                          )
				end if
			end if
		end if

		! ---------LOOP OVER INNER BLOCKS-------------------------------------------
		do iblock=1,ninner
			iinner = ilocal(iblock)                       ! Index in block of this CMP

			! Get cmpx/cmpy of cumulative CMP index; init modules
			call line2cart( (/ny1,ny2/), (iouter-1)*nblock+iinner, jcmp)
			icmp = grid2line2( (/ny1,ny2/), (/jcmp(1),jcmp(2)/) )
			if( narrow) call mnc_set_x02( x02_vect(icmp) )            ! Narrow azimuth

			call set_icmp( jcmp(1), jcmp(2))

			write(0,*) '  -----------------------------------------------------------'
			write(0,"(a,i2,a,i5,a,i5,a,i4)") '    thr=',ithr,' | CMPx=',jcmp(1),     &
			                                 ',CMPy=',jcmp(2),' | iblock=',iblock
			write(0,*) '  -----------------------------------------------------------'
			if( narrow) write(0,*) 'SET x02:',x02_vect(icmp)

			jd0 = grid2line2( (/ncmp,     nblock/), (/   1,     iblock/) )
			jd1 = grid2line2( (/ncmp,     nblock/), (/ncmp,     iblock/) )
			jm0 = grid2line3( (/ncmp,nmod,nblock/), (/   1,   1,iblock/) )
			jm1 = grid2line3( (/ncmp,nmod,nblock/), (/ncmp,nmod,iblock/) )


			! -- ONLY PROCESS DESIRED CMP RANGE --------------------------------------
			if(       (jcmp(1)>=iymin1) .AND. (jcmp(2)>=iymin2)                      &
			    .AND. (jcmp(1)<=iymax1) .AND. (jcmp(2)<=iymax2)  ) then

				if( modeling) then
					if(testing) call model_stats( 'FROM BLOCK', mblock(jm0:jm1), nmod,   &
					                               maxorder, nsurf, nt, nx1, nx2,        &
					                               jm0=jm0, ithr=ithr                    )
				else
					if(testing) call data_stats( 'FROM BLOCK', dblock(jd0:jd1),          &
					                             ithr=ithr, jd0=jd0, jd1=jd1             )
					if( optim) then
						call get_nullcube( nt, dt, t0, nx1, dx1, x01, nx2, dx2, x02, ny1,  &
						                   dy1, y01, ny2, dy2, y02, jcmp(1), jcmp(2),      &
						                   maxorder, nmod, nsurf, tauw_vect, vel, nullcube )

						call mnr_set_xtalk( xblock(jm0:jm1), nullcube )

						if(testing) call model_stats( 'CROSSTALK', xblock(jm0:jm1), nmod,  &
						                              maxorder, nsurf, nt, nx1, nx2,       &
						                              jm0=jm0, ithr=ithr                   )
					end if
				end if
	
				if(testing .AND. ndone==0 .AND. ithr==0 ) call mna_dp_test()    ! DPTest
				if(testing .AND. ndone==0 .AND. ithr==0 ) call mnc_dp_test()    ! DPTest
	
				! ------MAIN BLOCK -- OPTIMIZATION OR FORWARD OPER/ADJOINT--------------
				if( optim) then

					call get_reswt( nt, dt, t0, nx1, dx1, x01, nx2, dx2, x02, ny1, ny2,  &
					                jcmp(1), jcmp(2), nsurf, dblock(jd0:jd1), tauw_vect, &
					                vel, reswt                                           )
					if(testing) call data_stats( 'RESWT', reswt, ithr=ithr,  jd0=jd0,    &
					                             jd1=jd1                                 )
					call weight_init( reswt )

					model = mblock(jm0:jm1)
					data  = dblock(jd0:jd1)

					!---------------------------------------------------------------------
					! Starting guess for optimization: NMO, mute, stack, spread, distribute
					!---------------------------------------------------------------------
					if( .NOT. exist_file("m0") ) then
						stat = joint_nmo_row_exec( T, F, model, data )                   ! NMO
						im0 = grid2line2( (/ny1,ny2/), jcmp)
						it = nint( (tauw_vect(im0)-t0)/dt + 1 )
	
						do ix=1,nx1*nx2                                                 ! Mute
							im0 = grid2line2( (/nt,nx1*nx2/), (/it,nx1*nx2/) )
							im1 = grid2line2( (/nt,nx1*nx2/), (/nt,nx1*nx2/) )
							model(im0:im1) = 0.0
						end do
	
						do it=1,nt                                                     ! Stack
							im0 = grid2line2( (/nt,nx1*nx2/), (/it,1/) )
							nstk = 0
							do ix=1,nx1*nx2
								im1 = grid2line2( (/nt,nx1*nx2/), (/it,ix/) )
								model(im0) = model(im0) + model(im1)
								if( model(im1) /= 0.0 ) nstk = nstk+1
							end do
							if(nstk>0) model(im0) = model(im0)/nstk
						end do
	
						do it=1,nt                                                    ! Spread
							im0 = grid2line2( (/nt,nx1*nx2/), (/it,1/) )
							do ix=1,nx1*nx2
								im1 = grid2line2( (/nt,nx1*nx2/), (/it,ix/) )
								model(im1) = model(im0)
							end do
						end do
	
						do im=2,nmod                                              ! Distribute
							im0 = grid2line2( (/ncmp,nmod/), (/1,   im/) )
							im1 = grid2line2( (/ncmp,nmod/), (/ncmp,im/) )
							model(im0:im1) = model(1:ncmp)
							where( nullcube(im0:im1)==0 ) model(im0:im1) = 0.0
						end do
					end if
					!---------------------------------------------------------------------

					call solver_reg( model, data, joint_nmo_row_exec,                    &
					                 nmo_gradchain_exec, cgstep, 4*ncmp*nmod, niter,     &
					                 eps, verb=verb1, resd=rd, resm=rm, Wop=weight_lop,  &
					                 rmov=rmov, mmov=mmov, m0=model                      )

					mblock(jm0:jm1) = model
					dblock(jd0:jd1) = data

					if( jcmp(1)==cmpres(1) .AND. jcmp(2)==cmpres(2) ) then
						call sep_write( rd, "resd" )
						call sep_write( rm, "resm" )
						call sep_write( rmov(1:ncmp,:),  "rdmov" )
						call sep_write( rmov(ncmp+1:,:), "rmmov" )
						call sep_write( mmov, "mmov" )
						call tohistory_batch( "resd", nt, dt, t0, nx1, dx1, x01, n3=nx2,   &
                                  d3=dx2, o3=x02, n4=1, d4=1.0, o4=0.0, n5=1,  &
						                      d5=1.0, o5=0.0                               )
						call tohistory_batch( "resm", nt, dt, t0, nx1, dx1, x01, n3=nx2,   &
                                  d3=dx2, o3=x02, n4=nmod, d4=1.0, o4=0.0,     &
						                      n5=4, d5=1.0, o5=0.0, n6=1, d6=1.0, o6=0.0   )
						call tohistory_batch( "rdmov", nt, dt, t0, nx1, dx1, x01, n3=nx2,  &
						                      d3=dx2, o3=x02, n4=niter, d4=1.0, o4=0.0,    &
						                      n5=1, d5=1.0, o5=0.0, n6=1, d6=1.0, o6=0.0   )
						call tohistory_batch( "rmmov", nt, dt, t0, nx1, dx1, x01, n3=nx2,  &
						                      d3=dx2, o3=x02, n4=nmod, d4=1.0, o4=0.0,     &
						                      n5=4, d5=1.0, o5=0.0, n6=niter, d6=1.0,      &
						                      o6=0.0                                       )
						call tohistory_batch( "mmov", nt, dt, t0, nx1, dx1, x01, n3=nx2,   &
						                      d3=dx2, o3=x02, n4=nmod, d4=1.0, o4=0.0,     &
						                      n5=niter, d5=1.0, o5=0.0, n6=1, d6=1.0,      & 
						                      o6=0.0                                       )
					end if
  
					call cgstep_close()

				else
					stat = joint_nmo_row_exec( (.NOT. modeling), .false.,                &
					                           mblock(jm0:jm1), dblock(jd0:jd1)          )
				end if
				! ----------------------------------------------------------------------
	
				if(modeling) then
					if(testing) call data_stats( 'TO BLOCK2', dblock(jd0:jd1),           &
					                             ithr=ithr,  jd0=jd0, jd1=jd1            )
				else
					if(testing) call model_stats( 'TO BLOCK2', mblock(jm0:jm1), nmod,    &
					                              maxorder, nsurf, nt, nx1, nx2,         &
					                              ithr=ithr, jm0=jm0                     )
				end if

				ndone = ndone + 1

			end if

		! -----END LOOP OVER INNER BLOCKS-------------------------------------------
		end do

		if( modeling) then                                       ! Write data block
			call collect_write_main( "data", data_tag, dblock, ithr, nt, ncmp/nt,    &
			                         nbtmp, ninner, isend, test=testing              )
		else                                                     ! Write model block
			call collect_write_main( "model", model_tag, mblock, ithr, nt,           &
			                         ncmp*nmod/nt, nbtmp, ninner, isend, test=testing)
		end if

	end if !----(ninner > 0)------------------------------------------------------
end do
! -----END LOOP OVER OUTER BLOCKS-----------------------------------------------

call mult_nmo_core_close()
if(optim) call mult_nmo_reg_close()

deallocate( vel, tauw_vect, refl_vect, reswt, dip1, dip2, mblock, dblock,       &
            isend, rd                                                   )
if(optim) deallocate( xblock)

call mpi_finalize(stat)

call exit(0)

end program MultNmo3D_MPI
!--------------------------------------------------------------------------------
