!-------------------------------------------------------------------------------
! 2-D data transformation from offset to incidence angle squared, for AVO 
! analysis
!-------------------------------------------------------------------------------
program Offset2IncidenceAngle

use sep
use mult_nmo_util

implicit none

integer :: nt, nx, ny, j0, j1, itau, ix, iy, ia, ns
real    :: dt, dx, dy, t0, x0, y0, vw, x, vsq, w0, w1, smin, smax, ds, tau, sineang, xmax

real, dimension(:,:), allocatable :: data, model, vrms, fold

call sep_init(SOURCE)

call from_history('n1',nt)
call from_history('o1',t0)
call from_history('d1',dt)
call from_history('n2',nx)
call from_history('d2',dx)
call from_history('o2',x0)
call from_history('n3',ny)
call from_history('d3',dy)
call from_history('o3',y0)
call from_param('vw',vw,1500.)              ! Velocity at surface
call from_param('ns',ns,4*nx)              ! number of angles

allocate( vrms(nt,ny), data(nt,nx), fold(nt,ns), model(nt,ns) )

call sep_read( vrms,  'vrms')

vsq = vrms(1,1)**2
!smin = ( vw*(x0/vsq)*(1.0/sqrt( t0**2 + x0**2/vsq ) ) )**2
smin = 0.0
xmax = (nx-1)*dx + x0
smax = ( vw*(xmax/vsq)*(1.0/sqrt( t0**2 + xmax**2/vsq ) ) )**2
ds = (smax-smin)/(ns-1)

call to_history('n1',nt)
call to_history('d1',dt)
call to_history('o1',t0)
call to_history('n2',ns)
call to_history('d2',ds)
call to_history('o2',smin)
call to_history('n3',ny)
call to_history('d3',dy)
call to_history('o3',y0)

write(0,*) 'smin,smax,ds',smin,smax,ds

do iy=1,ny
	write(0,"(a,i4,a,i4)") 'done ',iy,' of ',ny

	call sep_read( data)

	model = 0.0
	fold = 0.0

	do itau=1,nt
		tau = (itau-1)*dt + t0
		vsq = vrms(itau,iy)**2

		do ix=1,nx
			x = (ix-1)*dx + x0
			sineang = ( vw*(x/vsq)/sqrt( tau**2 + x**2/vsq ) )**2
			call get_interp1_weights( sineang, ds, smin, j0, j1, w0, w1, 1, ns)
			model(itau,j0) = model(itau,j0) + w0*data(itau,ix)
			model(itau,j1) = model(itau,j1) + w1*data(itau,ix)
			fold(itau,j0) = fold(itau,j0) + w0
			fold(itau,j1) = fold(itau,j1) + w1
		end do
	end do

	where( fold>1.0e-4 ) model = model/fold

	call sep_write( model )
end do

deallocate( vrms, data, model, fold)

call exit(0)

end program Offset2IncidenceAngle
