! ---------------------------------------------------------------------------
!  RMSVelPick.f90 - Created by Morgan Brown
!           Wed Nov 28 17:36:59 2001.
! ---------------------------------------------------------------------------
!       Usage: RMSVelPick.x 
!  Parameters: 
! Description: 
! ---------------------------------------------------------------------------
program RMSVelPick

use sep

implicit none

integer  :: nt, nv, ny1, ny2, it, jmax(1), iy1, iy2, j0, j1, itauw, ewin
real     :: dt, dv, dy1, dy2, t0, v0, y01, y02, energy1, energy0, vl, vr,   &
            tau, taumax, tauw, vwmin, vwmax, vbmin, vbmax, vw, negdev, posdev
logical  :: searching

logical, parameter, save :: T=.true., F=.false.

real, dimension(:),     allocatable :: vprior 
real, dimension(:,:),   allocatable :: velin
real, dimension(:,:,:), allocatable :: velout, stackpow

call sep_init(SOURCE)

call from_history('n1',nt)  
call from_history('d1',dt)  
call from_history('o1',t0)  
call from_history('n2',nv)  
call from_history('d2',dv)  
call from_history('o2',v0)  
call from_history('n4',ny1)  
call from_history('d4',dy1)  
call from_history('o4',y01)  
call from_history('n5',ny2)  
call from_history('d5',dy2)  
call from_history('o5',y02)  

call from_param('vw',    vw,    1500. )
call from_param('vwmin', vwmin, 1400. )
call from_param('vwmax', vwmax, 1600. )
call from_param('vbmin', vbmin, 2000. )
call from_param('vbmax', vbmax, 3000. )

call to_history('n1',nt)  
call to_history('d1',dt)  
call to_history('o1',t0)  
call to_history('n2',ny1)  
call to_history('d2',dy1)  
call to_history('o2',y01)  
call to_history('n3',ny2)  
call to_history('d3',dy2)  
call to_history('o3',y02)  
call to_history('n4',1)  
call to_history('d4',1.0)  
call to_history('o4',0.0)  
call to_history('n5',1)  
call to_history('d5',1.0)  
call to_history('o5',0.0)  

call to_history('n1',nt,'stackpow')  
call to_history('d1',dt,'stackpow')  
call to_history('o1',t0,'stackpow')  
call to_history('n2',ny1,'stackpow')  
call to_history('d2',dy1,'stackpow')  
call to_history('o2',y01,'stackpow')  
call to_history('n3',ny2,'stackpow')  
call to_history('d3',dy2,'stackpow')  
call to_history('o3',y02,'stackpow')  
call to_history('n4',1,'stackpow')  
call to_history('d4',1.0,'stackpow')  
call to_history('o4',0.0,'stackpow')  
call to_history('n5',1,'stackpow')  
call to_history('d5',1.0,'stackpow')  
call to_history('o5',0.0,'stackpow')  

ewin = 5

allocate( velin(nt,nv), velout(nt,ny1,ny2), stackpow(nt,ny1,ny2), vprior(nt) )

if( exist_file("vprior") ) then
	call from_param('negdev', negdev, 0.01 )
	call from_param('posdev', posdev, 0.02 )
end if

do iy2=1,ny2
	do iy1=1,ny1
		call sep_read( velin)
		if( exist_file("vprior") ) then
			call sep_read( vprior, "vprior")
			do it=1,nt
				tau = (it-1)*dt + t0
				vl = vprior(it)*(1-negdev) 
				vr = vprior(it)*(1+posdev) 
				j0 = min(nv,max(1,nint( (vl-v0)/dv + 1 )))
				j1 = min(nv,max(1,nint( (vr-v0)/dv + 1 )))
				jmax = maxloc(velin(it,j0:j1)) + j0 - 1
				jmax = min(nv,max(1,jmax))
				velout(it,iy1,iy2) = (jmax(1)-1)*dv + v0
				stackpow(it,iy1,iy2) = velin(it,jmax(1))
			end do
		else
			j0 = nint( (vwmin-v0)/dv + 1 )
			j1 = nint( (vwmax-v0)/dv + 1 )
			energy0 = sum( velin(1:ewin,j0:j1) )	
			it = 1
			itauw = 1
			searching = T
			do while( it+ewin <= nt .AND. searching )    ! Detect WB reflection
				energy1 = sum( velin(it:it+ewin,j0:j1) )	
				if( energy1 > 50*energy0 ) then
					itauw = it
					searching = F
				end if
				it = it + 1
			end do
	
			taumax = (nt-1)*dt + t0
			tauw = (itauw-1)*dt + t0
			velout(1:itauw-1,iy1,iy2) = vw              ! Pad water V above WB refl
	
			write(0,*) 'cmpx/cmpy/tauw',iy1,iy2,tauw
			do it=itauw,nt
				tau = (it-1)*dt + t0
				vl = ( (vbmin-vwmin)/(taumax-tauw) )*(tau-tauw) + vwmin
				vr = ( (vbmax-vwmax)/(taumax-tauw) )*(tau-tauw) + vwmax
				j0 = min(nv,max(1,nint( (vl-v0)/dv + 1 )))
				j1 = min(nv,max(1,nint( (vr-v0)/dv + 1 )))
				jmax = maxloc(velin(it,j0:j1)) + j0 - 1
				jmax = min(nv,max(1,jmax))
				velout(it,iy1,iy2) = (jmax(1)-1)*dv + v0
				stackpow(it,iy1,iy2) = velin(it,jmax(1))
			end do		

		end if

	end do
end do

call sep_write( velout)
call sep_write( stackpow, "stackpow")

call exit(0)

end program RMSVelPick
