!-------------------------------------------------------------------------------
program RadonMultNAZ_MPI

use sep
use solver_reg_mod
use weight
use cgstep_mod
use mult_nmo_util
use mpi_sep
use mult_nmo_mpi
use cartesian
use hrt_narrowaz
use scaled_identity

implicit none

!-------------------------------------------------------------------------------
integer :: nt, nx1, nx2, ny1, ny2, stat, ncmp, ncmps, itest, iverb, maxmem,    &
           nblock, nbout, j0, j1, i0, jt0, jt1, iblock, jcmp(2), ithr, nthr,   &
           iouter, nouter, iinner, nbtmp, ninner, ilen, icmp, nv, iv, ix, it,  &
           it0, it1, iorder, k0, k1, niter
real    :: dt, dx1, dx2, dy1, dy2, t0, x01, x02, y01, y02, x2, dv, v0, v, x,   &
           tau, t, w0, w1, y1, y2, xp, eps
logical :: testing, verb1, verb2

integer, dimension(:), pointer :: isend, ilocal
real, dimension(:), pointer    :: radon, data, dblock, x02_vect, vel, vv, reswt
character(len=256) :: src, machine
character(len=20)  :: tmpstr
character(len=200) :: demult_tag, data_tag, x02_vect_tag, vel_tag

#ifdef SEP_MPI 
	src = SOURCE
	if( .not. mpi_sep_setup(src,ithr,nthr,verb2) )                               &
	    call seperr("MPI_SEP_SETUP(): Error")
	call MPI_get_processor_name(machine,ilen,stat)	
	write(0,"(a,i2,a,a)") 'SEP_MPI IS DEFINED, ithr=',ithr,', Machine=',         &
	                      trim(machine)
#else
	write(0,*) 'SEP_MPI NOT DEFINED, ithr=',ithr
	ithr=0;  nthr=1
#endif

call from_param( "maxmem" , maxmem, 200)   ! in MB
call from_param( "testing", itest , 0  ); testing=(itest==1)
call from_param( "verb"   , iverb , 0  ); verb1=(iverb>=1); verb2=(iverb>=2)

call mult_nmo_mpi_init(verb2)

!-------------------------------------------------------------------------------
call fromaux_batch( "data", nt, dt, t0, nx1, dx1, x01, n4=ny1, d4=dy1, o4=y01, &
                    n5=ny2, d5=dy2, o5=y02                                     )
nx2=1
nv = nx1                               ! Now, hard code nv=nx1 w/o (much) l.o.g.
call from_param( "v0", v0)
call from_param( "dv", dv)
call from_param( "niter", niter, 10)
call from_param( "eps", eps, 0.1)

call hrt_narrowaz_init( nt, nx1, nv, dt, dx1, dv, t0, x01, v0, verb=verb2)

if(ithr==0) call tohistory_batch( "demult", nt, dt, t0, nx1, dx1, x01, n3=1,   & 
                                  d3=1.0, o3=0.0, n4=ny1, d4=dy1, o4=y01,      &
                                  n5=ny2, d5=dy2, o5=y02                       )
ncmp = nt*nx1*nx2                                              ! Size of one CMP
ncmps = ny1*ny2                                                ! Total # of CMPs
nblock = nint( (maxmem*1000000)/(2.0*ncmp) )
nblock = min(nblock,ny1*ny2)

allocate( radon(ncmp), data(ncmp), dblock(ncmp*nblock), vv(nt), reswt(ncmp) )

!-------------------------------------------------------------------------------
! Get filetags for this thread, i.e., "model_tmp_1" and "dip_tmp_3"
!-------------------------------------------------------------------------------
tmpstr="demult";     call mpi_distrib_string( tmpstr, ithr, demult_tag     )
tmpstr="data";       call mpi_distrib_string( tmpstr, ithr, data_tag      )
tmpstr="x02_vect";   call mpi_distrib_string( tmpstr, ithr, x02_vect_tag  )
tmpstr="vel";        call mpi_distrib_string( tmpstr, ithr, vel_tag  )
! ------------------------------------------------------------------------------

! ------------------------------------------------------------------------------
! Read and distribute auxiliary files
! ------------------------------------------------------------------------------
allocate( x02_vect(ncmps), vel(nt*ncmp) )

call read_dist_aux( "x02_vect", x02_vect_tag, x02_vect, ithr, ncmps,           &
                    test=testing, check1=ny1, check2=ny2, default=x02          )
call read_dist_aux( "vel", vel_tag, vel, ithr, nt*ncmps, test=testing,         &
                    check1=nt, check2=ny1, check3=ny2                          )
! ------------------------------------------------------------------------------

if( nblock==ncmps ) then
	nouter = 1
else
	nouter = ncmps/nblock+1      ! integer math
end if

allocate( isend(nblock) )
if( .not. mpi_simple_distrib_pattern(isend,ilocal) ) call seperr("DIE DISTRIB")

! ---------LOOP OVER OUTER BLOCKS-----------------------------------------------
do iouter=1,nouter
	nbtmp = min(nblock,ncmps-(iouter-1)*nblock)       ! #CMPS in this outer block
	ninner = count(isend(1:nbtmp)==ithr)              ! #CMPS owned by this thread

	if( ninner>0 ) then
		write(0,*) '---------------------------------------------------------------'
		write(0,"(a,i2,a,i2,a,i2,a,i3,a,i3)") ' THREAD',ithr,': BLOCK ',&
		  iouter,' OF ',nouter,': BLOCKSZ=',nbtmp,'| NELEM in block=',ninner
		write(0,*) '---------------------------------------------------------------'
	
		call read_dist_main( "data", data_tag, dblock, ithr, ncmp, nbtmp, ninner,  & 
		                     isend, testing                                        )

		! ---------LOOP OVER INNER BLOCKS-------------------------------------------
		do iblock=1,ninner
			iinner = ilocal(iblock)                       ! Index in block of this CMP

			! Get cmpx/cmpy of cumulative CMP index; init modules
			call line2cart( (/ny1,ny2/), (iouter-1)*nblock+iinner, jcmp)
			icmp = grid2line2( (/ny1,ny2/), (/jcmp(1),jcmp(2)/) )

			write(0,*) '  -----------------------------------------------------------'
			write(0,"(a,i2,a,i5,a,i5,a,i4)") '    thr=',ithr,' | CMPx=',jcmp(1),     &
			                                 ',CMPy=',jcmp(2),' | iblock=',iblock
			write(0,*) '  -----------------------------------------------------------'

			j0 = grid2line2( (/ncmp,nblock/), (/1,   iblock/) )
			j1 = grid2line2( (/ncmp,nblock/), (/ncmp,iblock/) )
			k0 = grid2line3( (/nt,jcmp(1),jcmp(2)/), (/1, jcmp(1),jcmp(2)/) )
			k1 = grid2line3( (/nt,jcmp(1),jcmp(2)/), (/nt,jcmp(1),jcmp(2)/) )
			vv = vel(k0:k1)

			! ------MAIN BLOCK -------------------------------------------------------
			x2 = x02_vect(icmp)
			call set_x02( x2)
			y1 = (jcmp(1)-1)*dy1 + y01
			y2 = (jcmp(2)-1)*dy2 + y02

			reswt = 1.0
			do ix=1,nx1
				jt0 = j0 + grid2line2( (/nt,nx1/), (/1,ix/) ) - 1
				jt1 = j0 + grid2line2( (/nt,nx1/), (/nt,ix/) ) - 1
				k0 = grid2line2( (/nt,nx1/), (/1, ix/) )
				k1 = grid2line2( (/nt,nx1/), (/nt,ix/) )
				if( sum(abs(dblock(jt0:jt1))) < 1.0e-4 ) reswt(k0:k1) = 0.0
			end do

			write(0,*) 'min/max(data)',minval(dblock(j0:j1)),maxval(dblock(j0:j1))
			write(0,*) '#nonzero',nint(sum(reswt))

			if( nint(sum(reswt))==0 ) then
				write(0,*) 'skipping this gather--all zeros'
			else
				call weight_init( reswt )

				! LS Inversion
				call solver_reg( radon, dblock(j0:j1), hrt_narrowaz_exec, identity_exec, &
				                 cgstep, ncmp, niter, eps, verb=verb1, Wop=weight_lop    )
				call cgstep_close()
	
				! Mute Radon panel to estimate multiples
				do iv=1,nv
					v = (iv-1)*dv + v0
					do it=1,nt
						tau = (it-1)*dt + t0
						i0  = grid2line2( (/nt,nv/),  (/it,iv/)  )

						if( v >= 0.85*vv(it) .AND. v <= 0.9*vv(it) ) then
							radon(i0) = (0.9*vv(it) - v)/(0.05*vv(it))*radon(i0)
						else if( v > 0.9*vv(it) ) then
							radon(i0) = 0.0
						end if

!						if( v < 0.85*vv(it) .OR. v > 1.15*vv(it) ) then
!							radon(i0) = 0.0
!						else if( v >= 0.85*vv(it) .AND. v <= 0.9*vv(it) ) then
!							radon(i0) = (v - 0.85*vv(it))/(0.05*vv(it))*radon(i0)
!						else if( v >= 1.1*vv(it) .AND. v <= 1.15*vv(it) ) then
!							radon(i0) = (1.15*vv(it) - v)/(0.05*vv(it))*radon(i0)
!						end if
					end do
				end do

				! forward model muted radon panel = multiples
				stat = hrt_narrowaz_exec( .false., .false., radon, data)

				! subtract estimated multiples from data	
				dblock(j0:j1) = dblock(j0:j1) - data

			end if
			! ------------------------------------------------------------------------
	
		! -----END LOOP OVER INNER BLOCKS-------------------------------------------
		end do

		call collect_write_main( "demult", demult_tag, dblock, ithr, nt, ncmp/nt,  &
		                         nbtmp, ninner, isend, test=testing                )

	end if !----(ninner > 0)------------------------------------------------------
end do
! -----END LOOP OVER OUTER BLOCKS-----------------------------------------------

deallocate( dblock, vel, vv, isend, x02_vect, reswt )

call mpi_finalize(stat)

call exit(0)

end program RadonMultNAZ_MPI
