!-------------------------------------------------------------------------------
program VelanNarrowAzimuth_MPI

use sep
use mult_nmo_util
use morgan_util
use mpi_sep
use mult_nmo_mpi
use cartesian

implicit none

!-------------------------------------------------------------------------------
integer :: nt, nx1, nx2, ny1, ny2, stat, ncmp, ncmps, itest, iverb, maxmem,    &
           nblock, nbout, j0, j1, i0, jt0, jt1, iblock, jcmp(2), ithr, nthr,   &
           iouter, nouter, iinner, nbtmp, ninner, ilen, icmp, nv, iv, ix, it,  &
           it0, it1, iorder
real    :: dt, dx1, dx2, dy1, dy2, t0, x01, x02, y01, y02, x2, dv, v0, v, x,   &
           tau, t, w0, w1, tauw, y1, y2, xp
logical :: testing, verb

integer, dimension(:), pointer :: isend, ilocal
real, dimension(:), pointer    :: mblock, dblock, x02_vect, tauw_vect
character(len=256) :: src, machine
character(len=20)  :: tmpstr
character(len=200) :: model_tag, data_tag, x02_vect_tag, tauw_vect_tag

#ifdef SEP_MPI 
	src = SOURCE
	if( .not. mpi_sep_setup(src,ithr,nthr,verb) )                                &
	    call seperr("MPI_SEP_SETUP(): Error")
	call MPI_get_processor_name(machine,ilen,stat)	
	write(0,"(a,i2,a,a)") 'SEP_MPI IS DEFINED, ithr=',ithr,', Machine=',         &
	                      trim(machine)
#else
	write(0,*) 'SEP_MPI NOT DEFINED, ithr=',ithr
	ithr=0;  nthr=1
#endif

call from_param( "maxmem" , maxmem, 200)   ! in MB
call from_param( "testing", itest , 0  ); testing=(itest==1)
call from_param( "verb"   , iverb , 0  ); verb=(iverb==1)
call from_param( "order"  , iorder, 0  )

call mult_nmo_mpi_init(verb)

!-------------------------------------------------------------------------------
call fromaux_batch( "data", nt, dt, t0, nx1, dx1, x01, n4=ny1, d4=dy1, o4=y01, &
                    n5=ny2, d5=dy2, o5=y02                                     )

nx2=1
nv = nx1                               ! Now, hard code nv=nx1 w/o (much) l.o.g.
call from_param( "v0", v0)
call from_param( "dv", dv)

if(ithr==0) call tohistory_batch( "model", nt, dt, t0, nv, dv, v0, n3=1,       & 
                                  d3=1.0, o3=0.0, n4=ny1, d4=dy1, o4=y01,      &
                                  n5=ny2, d5=dy2, o5=y02                       )
ncmp = nt*nx1*nx2                                              ! Size of one CMP
ncmps = ny1*ny2                                                ! Total # of CMPs
nblock = nint( (maxmem*1000000)/(2.0*ncmp) )
nblock = min(nblock,ny1*ny2)

allocate( mblock(ncmp*nblock), dblock(ncmp*nblock) )

!-------------------------------------------------------------------------------
! Get filetags for this thread, i.e., "model_tmp_1" and "dip_tmp_3"
!-------------------------------------------------------------------------------
tmpstr="model";      call mpi_distrib_string( tmpstr, ithr, model_tag     )
tmpstr="data";       call mpi_distrib_string( tmpstr, ithr, data_tag      )
tmpstr="x02_vect";   call mpi_distrib_string( tmpstr, ithr, x02_vect_tag  )
tmpstr="tauw_vect";   call mpi_distrib_string( tmpstr, ithr, tauw_vect_tag  )
! ------------------------------------------------------------------------------

! ------------------------------------------------------------------------------
! Read and distribute auxiliary files
! ------------------------------------------------------------------------------
allocate( x02_vect(ncmps), tauw_vect(ncmps) )

call read_dist_aux( "x02_vect", x02_vect_tag, x02_vect, ithr, ncmps,           &
                    test=testing, check1=ny1, check2=ny2, default=x02          )
call read_dist_aux( "tauw_vect", tauw_vect_tag, tauw_vect, ithr, ncmps,        &
                    test=testing, check1=ny1, check2=ny2, default=0.0          )
! ------------------------------------------------------------------------------

if( nblock==ncmps ) then
	nouter = 1
else
	nouter = ncmps/nblock+1      ! integer math
end if

allocate( isend(nblock) )
if( .not. mpi_simple_distrib_pattern(isend,ilocal) ) call seperr("DIE DISTRIB")

! ---------LOOP OVER OUTER BLOCKS-----------------------------------------------
do iouter=1,nouter
	nbtmp = min(nblock,ncmps-(iouter-1)*nblock)       ! #CMPS in this outer block
	ninner = count(isend(1:nbtmp)==ithr)              ! #CMPS owned by this thread

	if( ninner>0 ) then
		write(0,*) '---------------------------------------------------------------'
		write(0,"(a,i2,a,i2,a,i2,a,i3,a,i3)") ' THREAD',ithr,': BLOCK ',&
		  iouter,' OF ',nouter,': BLOCKSZ=',nbtmp,'| NELEM in block=',ninner
		write(0,*) '---------------------------------------------------------------'
	
		call read_dist_main( "data", data_tag, dblock, ithr, ncmp, nbtmp, ninner,  & 
		                     isend, testing                                        )

		! ---------LOOP OVER INNER BLOCKS-------------------------------------------
		do iblock=1,ninner
			iinner = ilocal(iblock)                       ! Index in block of this CMP

			! Get cmpx/cmpy of cumulative CMP index; init modules
			call line2cart( (/ny1,ny2/), (iouter-1)*nblock+iinner, jcmp)
			icmp = grid2line2( (/ny1,ny2/), (/jcmp(1),jcmp(2)/) )


			write(0,*) '  -----------------------------------------------------------'
			write(0,"(a,i2,a,i5,a,i5,a,i4)") '    thr=',ithr,' | CMPx=',jcmp(1),     &
			                                 ',CMPy=',jcmp(2),' | iblock=',iblock
			write(0,*) '  -----------------------------------------------------------'

			j0 = grid2line2( (/ncmp,nblock/), (/1,   iblock/) )
			j1 = grid2line2( (/ncmp,nblock/), (/ncmp,iblock/) )

			mblock(j0:j1) = 0.0
	
			! ------MAIN BLOCK -------------------------------------------------------
			x2 = x02_vect(icmp)
			tauw = tauw_vect(icmp)
			y1 = (jcmp(1)-1)*dy1 + y01
			y2 = (jcmp(2)-1)*dy2 + y02

			do iv=1,nv
				v = (iv-1)*dv + v0
				do ix=1,nx1
					x = (ix-1)*dx1 + x01
					do it=1,nt
						tau = ( (it-1)*dt + t0 )*(iorder+1)

!						if( iorder==0 ) then                             ! Normal velan
!							tau = (it-1)*dt + t0
!						else if( iorder==1 ) then                        ! Special WBM velan
!							xp = sqrt( x*x + x2*x2 )
!							tau = get_taumult( y1, y2, x, x2, xp, ny1, ny2, dy1, dy2, y01,   &
!							                   y02, tauw_vect, 1, 0, 1, 1 )                  &
!							    + get_taumult( y1, y2, x, x2, xp, ny1, ny2, dy1, dy2, y01,   &
!							                   y02, tauw_vect, 1, 1, 1, 1 )
!						else if( iorder>1 ) then
!							call seperr("Order > 1 not yet implemented")
!						end if

						t = sqrt( tau**2 + (x2*x2+x*x)/(v*v) )

						call get_interp1_weights( t, dt, t0, it0, it1, w0, w1, 1, nt)

						i0  = j0 + grid2line2( (/nt,nv/),  (/it,iv/)  ) - 1
						jt0 = j0 + grid2line2( (/nt,nx1/), (/it0,ix/) ) - 1
						jt1 = j0 + grid2line2( (/nt,nx1/), (/it1,ix/) ) - 1

						mblock(i0) = mblock(i0) + w0*dblock(jt0) + w1*dblock(jt1)

					end do
				end do
			end do
			! ------------------------------------------------------------------------
	
		! -----END LOOP OVER INNER BLOCKS-------------------------------------------
		end do

		call collect_write_main( "model", model_tag, mblock, ithr, nt, ncmp/nt,    &
		                         nbtmp, ninner, isend, test=testing                )

	end if !----(ninner > 0)------------------------------------------------------
end do
! -----END LOOP OVER OUTER BLOCKS-----------------------------------------------

deallocate( mblock, dblock, isend, x02_vect )

call mpi_finalize(stat)

call exit(0)

end program VelanNarrowAzimuth_MPI
