!-------------------------------------------------------------------------------
! the first live trace on adjacent CMP gathers may be different.  This is 
! equivalent to windowing with a variable f2
!-------------------------------------------------------------------------------
program WindFirstTrace

use sep

implicit none

integer :: nt, nx1, nx2, ny1, ny2, ntwind, j0, j1, ix1, ix2, iy1, iy2, j2, f2min, f2max, n2out, f2
real    :: dt, dx1, dx2, dy1, dy2, t0, x01, x02, y01, y02, o2out, f2real(1)
logical :: found

logical, parameter :: T=.true., F=.false.

real, dimension(:,:,:)   :: data, model

call sep_init(SOURCE)

call from_history('n1',nt)
call from_history('o1',t0)
call from_history('d1',dt)
call from_history('n2',nx1)
call from_history('d2',dx1)
call from_history('o2',x01)
call from_history('n3',nx2)
call from_history('d3',dx2)
call from_history('o3',x02)
call from_history('n4',ny1)
call from_history('d4',dy1)
call from_history('o4',y01)
call from_history('n5',ny2)
call from_history('d5',dy2)
call from_history('o5',y02)

call from_param('j2',j2,1)
call from_param('f2min',f2min,1)
call from_param('f2max',f2max,j2)

n2out = nint(1.0*nx1/j2)
o2out = (f2min-1)*dx1 + x01

allocate( data(nt,nx1,nx2), model(nt,n2out,nx2) )

do iy2=1,ny2
	do iy1=1,ny1
		call sep_read( data)	
		write(0,"(a,2i4,a,2i4)") 'done ',iy1,iy2,' of ',ny1,ny2

		do ix2=1,nx2

			if( exist_file("f2in") ) then
				call sep_read( f2real(1:1), "f2in")
				f2 = nint(f2real(1))
			else
				found = F
				f2 = 1
				do ix1=f2min,f2max
					if( sum(abs(data(:,ix1,ix2))) /= 0.0 .AND. (.NOT. found) ) then
						f2 = ix1
						found = T
					end if
				end do
			end if

			j0 = 1
			do ix1=f2,nx1,j2
				model(:,j0,ix2) = data(:,ix1,ix2)
				j0 = j0 + 1
			end do
		end do

		if( exist_file("f2out") ) then
			f2real(1) = 1.0*f2
			call sep_write( f2real(1:1), "f2out")
		end if

		call sep_write(model)

	end do
end do

call to_history('n1',nt)
call to_history('d1',dt)
call to_history('o1',t0)
call to_history('n2',n2out)
call to_history('d2',dx1)
call to_history('o2',o2out)
call to_history('n3',nx2)
call to_history('d3',dx2)
call to_history('o3',x02)
call to_history('n4',ny1)
call to_history('d4',dy1)
call to_history('o4',y01)
call to_history('n5',ny2)
call to_history('d5',dy2)
call to_history('o5',y02)

if( exist_file("f2out") ) then
	call to_history('n1',ny1,'f2out')
	call to_history('d1',dy1,'f2out')
	call to_history('o1',y01,'f2out')
	call to_history('n2',ny2,'f2out')
	call to_history('d2',dy2,'f2out')
	call to_history('o2',y02,'f2out')
	call to_history('n3',1,'f2out')
	call to_history('n4',1,'f2out')
	call to_history('n5',1,'f2out')
end if

deallocate( data, model)

call exit(0)

end program WindFirstTrace
