!-------------------------------------------------------------------------------
!  hrt_narrowaz.f90 - Created by Morgan Brown
!           Sun Apr 11 12:23:41 2004.
!-------------------------------------------------------------------------------
!       Usage: hrt_narrowaz.mod 
!  Parameters: 
! Description: 
!-------------------------------------------------------------------------------
module hrt_narrowaz

use sep
use adj_mod
use dottest
use morgan_util
use mult_nmo_util

implicit none

integer, private, save :: nt, nx1, nv
real,    private, save :: dt, dx1, t0, x01, v0, dv, x02
logical, private, save :: verbose
logical, parameter, private, save :: T=.true., F=.false.
real, dimension(:), pointer, private, save :: tmp

contains

!-------------------------------------------------------------------------------
subroutine hrt_narrowaz_init( nt_in, nx1_in, nv_in, dt_in, dx1_in, dv_in,      &
                              t0_in, x01_in, v0_in, verb )
	integer, intent(IN) :: nt_in, nx1_in, nv_in
	real,    intent(IN) :: dt_in, dx1_in, t0_in, x01_in, v0_in, dv_in
	logical  :: verb
	optional :: verb

	nt = nt_in
	nx1 = nx1_in
	nv = nv_in
	dt = dt_in
	dx1 = dx1_in
	dv = dv_in
	t0 = t0_in
	x01 = x01_in
	v0 = v0_in

	verbose=F;  if(present(verb)) verbose=verb
end subroutine hrt_narrowaz_init
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
subroutine hrt_narrowaz_close( )
end subroutine hrt_narrowaz_close
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
subroutine set_x02( x02_in)
	real, intent(IN) :: x02_in

	x02 = x02_in
end subroutine set_x02
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
integer function hrt_narrowaz_exec( adj, add, model, data)
	logical            :: adj, add 
	real, dimension(:) :: model, data 

	integer :: iv, ix, it, i0, jt0, jt1, it0, it1
	real    :: v, x, tau, tt, w0, w1

	call adjnull( adj, add, model, data)

	if( verbose) call oper_stats( 'hrt_narrowaz_exec', T, model, data, adj)

	do iv=1,nv
		v = (iv-1)*dv + v0
		do ix=1,nx1
			x = (ix-1)*dx1 + x01
			do it=1,nt
				tau = (it-1)*dt + t0

				tt = sqrt( tau**2 + (x02*x02+x*x)/(v*v) )

				call get_interp1_weights( tt, dt, t0, it0, it1, w0, w1, 1, nt)

				i0  = grid2line2( (/nt,nv/),  (/it,iv/)  )
				jt0 = grid2line2( (/nt,nx1/), (/it0,ix/) )
				jt1 = grid2line2( (/nt,nx1/), (/it1,ix/) )

				if( adj) then
					model(i0) = model(i0) + w0*data(jt0) + w1*data(jt1)
				else
					data(jt0) = data(jt0) + w0*model(i0)
					data(jt1) = data(jt1) + w1*model(i0)
				end if
			end do
		end do
	end do

	if( verbose) call oper_stats( 'hrt_narrowaz_exec', F, model, data, adj)

	hrt_narrowaz_exec = 0

end function hrt_narrowaz_exec
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
!subroutine hrt_narrowaz_test( )
!	real    :: dot1(2), dot2(2)
!	integer :: nmod, ndat
!
!	call hrt_narrowaz_init( )
!	call dot_test( hrt_narrowaz_exec, nmod, ndat, dot1, dot2)
!end subroutine hrt_narrowaz_test
!-------------------------------------------------------------------------------

end module hrt_narrowaz
