!--------------------------------------------------------------------------------
! ChangeLog:
!--------------------------------------------------------------------------------
! 9 Oct 2003: Normalized Snell was causing all hell to break loose.  Added option
!             (normsnell=T/F) when normalized snell is needed.
!--------------------------------------------------------------------------------
module mult_nmo_ampl_3d

use adj_mod
use mult_nmo_util
use dottest
use scaled_identity
use chain0_mod
use weight2

implicit none

integer, private, save  :: nt, nx1, nx2, ny1, ny2, korder, iy1, iy2, isplit,& 
                           ithr, ncmp, ncmps, isurf, nsurf, nmod
real,    private, save  :: dt, t0, dx1, x01, dx2, x02, dy1, y01, dy2, y02
logical, private, save  :: verbose, gscinv, test, reflinv, narrow_az, gsc,     &
                           snell, snellnorm

real, dimension(:), private, pointer, save :: v, refl, tauw, refl_tab,         &
                                              snellwt, tmpsnellwt, tmp
logical, dimension(:,:), private, pointer, save :: misstrace

logical, private, parameter, save :: T = .true., F = .false.

contains

!--------------------------------------------------------------------------------
subroutine mult_nmo_ampl_init( vel, tauw_in, nt_in, t0_in, dt_in, nx1_in,       &
                               x01_in, dx1_in, nx2_in, x02_in, dx2_in, ny1_in,  &
                               y01_in, dy1_in, ny2_in, y02_in, dy2_in,          &
                               nsurf_in, refl_in, korder_in, nmod_in, verb,     &
                               invgsc, testing, invrefl, ithread, narrow,       &
                               do_gsc, do_snell, normsnell                      )

	real,    intent(IN) :: x01_in, dx1_in, x02_in, dx2_in, y01_in, dy1_in,        &
	                       y02_in, dy2_in, t0_in, dt_in
	integer, intent(IN) :: nx1_in, nx2_in, ny1_in, ny2_in, korder_in, nt_in,      &
	                       nsurf_in, nmod_in
	integer :: ithread
	logical :: verb, invgsc, testing, invrefl, narrow, do_gsc, do_snell, normsnell
	real, dimension(:), pointer :: vel, tauw_in, refl_in

	optional :: verb, invgsc, testing, invrefl, ithread, narrow, do_gsc,          &
	            do_snell, normsnell

	if(verb) write(0,*) 'mult_nmo_ampl_init(): TOP'

	nt  = nt_in;    t0 = t0_in;   dt = dt_in               ! Time axis
	nx1 = nx1_in;  x01 = x01_in;  dx1 = dx1_in             ! Offset-x axis
	nx2 = nx2_in;  x02 = x02_in;  dx2 = dx2_in             ! Offset-y axis
	ny1 = ny1_in;  y01 = y01_in;  dy1 = dy1_in             ! Midpoint-x axis
	ny2 = ny2_in;  y02 = y02_in;  dy2 = dy2_in             ! Midpoint-y axis
	nsurf = nsurf_in
	nmod = nmod_in

	ncmp  = nt*nx1*nx2
	ncmps = ny1*ny2

	call mna_set_order(korder_in,0,1)

	allocate( tauw(ncmps*nsurf), refl(ncmps*nsurf), v(nt*ncmps), tmp(ncmp) )
	allocate( refl_tab(ncmp*nmod), snellwt(ncmp*nmod), tmpsnellwt(ncmp) )
	allocate( misstrace(nx1,nx2) )
	tauw = tauw_in;  refl = refl_in;  v = vel
	snellwt = 1.0
	misstrace = F

	verbose   = F ; if( present(verb)      ) verbose   = verb
	gscinv    = F ; if( present(invgsc)    ) gscinv    = invgsc
	reflinv   = F ; if( present(invrefl)   ) reflinv   = invrefl
	test      = F ; if( present(testing)   ) test      = testing
	gsc       = T ; if( present(do_gsc)    ) gsc       = do_gsc
	snell     = T ; if( present(do_snell)  ) snell     = do_snell
	test      = F ; if( present(testing)   ) test      = testing
	snellnorm = F ; if( present(normsnell) ) snellnorm = normsnell
	ithr      = 0 ; if( present(ithread)   ) ithr      = ithread
	narrow_az = F ; if( present(narrow)    ) narrow_az = narrow

	if(verbose) write(0,*) 'mult_nmo_ampl_init(): BOT'

end subroutine mult_nmo_ampl_init
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
subroutine mna_set_icmp( icmp1,icmp2,norder,tracemiss)
	integer :: icmp1, icmp2, norder
	logical, dimension(:,:) :: tracemiss
	optional                :: tracemiss
	if(verbose) write(0,*) 'mna_set_icmp(): TOP'
	iy1=icmp1
	iy2=icmp2
	if(present(tracemiss)) misstrace = tracemiss
	call tabulate_ampl(norder)
	if(verbose) write(0,*) 'mna_set_icmp(): BOT'
end subroutine mna_set_icmp
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
subroutine mna_set_order(ord,spl,surf)
	integer, intent(IN) :: ord, spl, surf
	if(verbose) write(0,*) 'mna_set_order(): TOP'
	korder = ord
	isplit = spl
	isurf = surf
	if(verbose) write(0,*) 'mna_set_order(): BOT'
end subroutine mna_set_order
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
	subroutine mult_nmo_ampl_close( )
		if(verbose) write(0,*) 'mult_nmo_ampl_close(): TOP'
		if( test) then
			if( .NOT. allocated(v)    ) call seperr( 'ERROR! v not allocated'   )
			if( .NOT. allocated(tauw) ) call seperr( 'ERROR! tauw not allocated')
			if( .NOT. allocated(refl) ) call seperr( 'ERROR! refl not allocated')
			if( .NOT. allocated(refl_tab) )                                          &
				call seperr( 'ERROR! refl_tab not allocated')
			if( .NOT. allocated(snellwt) )                                           &
				call seperr( 'ERROR! snellwt not allocated')
			if( .NOT. allocated(tmpsnellwt) )                                        &
				call seperr( 'ERROR! tmpsnellwt not allocated')
			if( .NOT. allocated(misstrace) )                                         &
				call seperr( 'ERROR! misstrace not allocated')
		end if
		deallocate( v, tauw, refl, refl_tab, snellwt, tmpsnellwt, misstrace)
		if(verbose) write(0,*) 'mult_nmo_ampl_close(): BOT'
	end subroutine mult_nmo_ampl_close
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
! Relative geometric spreading correction for pegleg of order k.  Follows
! Ursin (1990), and more specifically, Lu et al. (1999).
!
!  g_prim(x) = t_nmo_prim(x)*v_water;  g_mult(x) = t_nmo_mult(x)*v_water
!
! Apply ratio [g_mult/g_prim](x) to model vector.
!--------------------------------------------------------------------------------
integer function geometric_spreading_exec( adj, add, model, data)
	integer                     :: itau, ix1, ix2, k, stat, jv, iy, itauw
	real                        :: tau, ttauk, vveff, x1, x2, xtot, g, gprim,     &
	                               gmult, vtmp
	logical, intent(IN)         :: adj, add 
	real, dimension(:)          :: model, data 

	call adjnull( adj, add, model, data)
	if(verbose) call oper_stats( 'geometric_spreading_exec():',.true., model,    &
	                             data, adj, ithr=ithr                            )

	if( korder==0 .OR. (.NOT. gsc) ) then
		stat = identity_exec(adj,add,model,data)
	else
		iy = grid2line3( (/ny1,ny2,nsurf/), (/iy1,iy2,isurf/) )
		itauw = grid2line3( (/nt,ny1,ny2/), (/nint((tauw(iy)-t0)/dt+1),iy1,iy2/) )
		do itau = 1,nt
			tau = t0 + (itau-1)*dt
			ttauk = tauw(iy)*korder + tau
			jv = grid2line3( (/nt,ny1,ny2/), (/itau,iy1,iy2/) )
			vveff = (tauw(iy)*v(itauw)*v(itauw)*korder + tau*v(jv)*v(jv))/ttauk
			do ix2 = 1,nx2
				x2 = (ix2-1)*dx2 + x02	
				do ix1 = 1,nx1
					x1 = (ix1-1)*dx1 + x01	
					xtot = sqrt( x1*x1 + x2*x2)
					gprim = sqrt( tau**2 + (xtot/v(jv))**2 )
					gmult = sqrt( ttauk**2 + xtot*xtot/vveff )
	
					if( gscinv ) then
						g=1
						if( gprim > 1.0e-6)  g = gmult/gprim
					else
						g = gprim/gmult
					end if

					k = grid2line3( (/nt,nx1,nx2/), (/itau,ix1,ix2/) )
					if( adj) then
						model(k) = model(k) + data(k)*g
					else
						data(k) = data(k) + model(k)*g
					end if
				end do
			end do
		end do
	end if

	if(verbose) call oper_stats( 'geometric_spreading_exec():', .false., model,  &
	                             data, adj, ithr=ithr                            )
	geometric_spreading_exec = 0
end function geometric_spreading_exec
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
! Applied to model space panel AFTER NMO'.  Mutes area where energy is
! nonsensical, i.e., before onset of first WB multiple on panel m_1.
!
! Also applies reflection coefficient to panel.  
!--------------------------------------------------------------------------------
integer function onset_mute_exec( adj, add, model, data)
	integer                     :: ix1, ix2, it, j0, itmute, thresh, iy, itauw,  &
	                               k, j
	real                        :: x1, x2, xtot, tmute, rr, xp, y1, y2
	logical, intent(IN)         :: adj, add
	real, dimension(:)          :: model, data


	call adjnull( adj, add, model, data)
	if(verbose) call oper_stats( 'onset_mute_exec()', .true., model, data, adj,  &
	                             ithr=ithr                                       )
	thresh = 10
	y1 = (iy1-1)*dy1+y01;  y2 = (iy2-1)*dy2+y02
	iy = grid2line3( (/ny1,ny2,nsurf/), (/iy1,iy2,isurf/) )
	itauw = grid2line3( (/nt,ny1,ny2/), (/nint((tauw(iy)-t0)/dt+1),iy1,iy2/) )

	do ix2=1,nx2
		x2 = (ix2-1)*dx2 + x02
		do ix1=1,nx1
			x1 = (ix1-1)*dx1 + x01
			xtot = sqrt( x1*x1 + x2*x2 )
			tmute = sqrt( ((korder+1)*tauw(iy))**2 + (xtot/v(itauw))**2 )
			itmute = min( nt, max(1,nint( (tmute-t0)/dt - thresh )) )
			do it=itmute,nt
				k = get_cubeloc( korder, nsurf, isurf, isplit )
				j = grid2line4( (/nt,nx1,nx2,nmod/), (/it,ix1,ix2,k/) )
				rr = refl_tab(j)
	
				j0 = grid2line3( (/nt,nx1,nx2/), (/it,ix1,ix2/) )
				if( adj) then
					model(j0) = model(j0) + data(j0)*rr
				else
					data(j0)  = data(j0) + model(j0)*rr
				end if
			end do
		end do
	end do

	if(verbose) call oper_stats( 'onset_mute_exec()', .false., model, data, adj, &
	                             ithr=ithr                                       )
	onset_mute_exec = 0
end function onset_mute_exec
! ------------------------------------------------------------------------------

! ------------------------------------------------------------------------------
! Does Snell resampling and normalizes to compensate for x-axis compression.
! ------------------------------------------------------------------------------
integer function snell_resamp_exec( adj, add, model, data)
	integer                     :: k, j0, j1, stat
	logical, intent(IN)         :: adj, add
	real, dimension(:)          :: model, data

	call adjnull( adj, add, model, data)

	if(verbose) call oper_stats( 'snell_resamp_exec()', .true., model, data,     &
	                             adj, ithr=ithr                                  )

	if( korder==0 .OR. (.NOT. snell) ) then
		stat = identity_exec( adj, add, model, data)
	else
		if( snellnorm) then
			k = get_cubeloc( korder, nsurf, isurf, isplit)
			j0 = grid2line2( (/ncmp,nmod/), (/1,k/) )
			j1 = grid2line2( (/ncmp,nmod/), (/ncmp,k/) )
			tmpsnellwt = snellwt(j0:j1)
			call weight2_init( tmpsnellwt)
			call chain0( snell_resamp_core_exec, weight2_lop, adj, add, model,      &
			             data, tmp                                                  )
		else
			stat = snell_resamp_core_exec( adj, add, model, data)
		end if
	end if

	if(verbose) call oper_stats( 'snell_resamp_exec()', .false., model, data,   &
	                             adj, ithr=ithr                                 )

	snell_resamp_exec = 0
end function snell_resamp_exec
! ------------------------------------------------------------------------------


! ------------------------------------------------------------------------------
! Loops over data (normal) space.  Model space is compressed.  Compresses an 
! i^th order multiple to an equivalent primary.    
!
!    xp^2 = xm*tau*V^2/( ((tauw*i+tau)*Ve^2)^2 + xm^2(Ve^2-V^2) )
!
! In constant velocity, Ve = V, and this reduces to a radial trace-like
! resampling (Taner).
! ------------------------------------------------------------------------------
integer function snell_resamp_core_exec( adj, add, model, data)
	integer                     :: i0, im00, im01, im10, im11, it, ix1, ix2,     &
	                               ixhi(2), ixlo(2), stat, jv, iy, im(4), i, k,  &
	                               itauw, j0, j1
	real                        :: x1, x2, tt, xp, vv, wlo(2), whi(2), ttauk,    &
	                               xp1, xp2, w(4)
	logical, intent(IN)         :: adj, add
	real, dimension(:)          :: model, data

	call adjnull( adj, add, model, data)

	if(verbose) call oper_stats( 'snell_resamp_core_exec()', .true., model, data, &
	                             adj, ithr=ithr                                   )

	k = get_cubeloc( korder, nsurf, isurf, isplit)
	j0 = grid2line2( (/ncmp,nmod/), (/1,k/) )
	j1 = grid2line2( (/ncmp,nmod/), (/ncmp,k/) )
	
	iy = grid2line3( (/ny1,ny2,nsurf/), (/iy1,iy2,isurf/) )
	itauw = grid2line3( (/nt,ny1,ny2/), (/nint((tauw(iy)-t0)/dt+1),iy1,iy2/) )

	if( korder==0 ) then
		stat = identity_exec(adj,add,model,data)
	else
		do it=1,nt              ! Loop over data space
			tt = (it-1)*dt + t0
			jv = grid2line3( (/nt,ny1,ny2/), (/it,iy1,iy2/) )
			do ix2=1,nx2
				x2 = (ix2-1)*dx2 + x02
				do ix1=1,nx1
					x1 = (ix1-1)*dx1 + x01
					xp = get_xp( x1, tt, v(jv), tauw(iy), korder, v(itauw), x2=x2 )
					call get_xp12( xp, x1, x2, xp1, xp2)
					if( narrow_az ) xp2 = x2
					if( nx2==1 ) then
						call get_interp1_weights( xp1, dx1, x01, ixlo(1), ixhi(1), wlo(1),   &
						                          whi(1), 1, nx1                             )
						ixlo(2) = 1;   ixhi(2) = 1
						 wlo(2) = 0.5;  whi(2) = 0.5
					else
						call get_interpn_weights( (/xp1,xp2/), (/dx1,dx2/), (/x01,x02/),     &
						                          ixlo, ixhi, wlo, whi, (/1,1/), (/nx1,nx2/) )
					end if
					call bins_weights_to_1d( (/nx1,nx2/), wlo, whi, ixlo, ixhi, w, im)
					i0 = grid2line3( (/nt,nx1,nx2/), (/it,ix1,ix2/) )
					do i=1,4
						k = grid2line2( (/nt,nx1*nx2/), (/it,im(i)/) )
						if(adj) then
							model(k) = model(k) + w(i)*data(i0)
						else
							data(i0) = data(i0) + model(k)*w(i)
						end if
					end do
				end do
			end do
		end do

	end if

	if(verbose) call oper_stats( 'snell_resamp_core_exec()', .false., model, data, &
	                             adj, ithr=ithr                                    )
	snell_resamp_core_exec = 0
end function snell_resamp_core_exec
! ------------------------------------------------------------------------------

! ------------------------------------------------------------------------------
subroutine tabulate_ampl(norder)
	integer, intent(IN) :: norder
	integer             :: ix1, ix2, it, itmute, thresh, iy, itauw, jt, j,       &
	                       nnsurf, jorder, jsurf, jsplit, j0, j1, k, stat
	real                :: x1, x2, xtot, tmute, rr, tt, xp, y1, y2
	real, dimension(:), allocatable :: d_ones

	if(verbose) write(0,*) 'tabulate_ampl(): TOP'

	y1 = (iy1-1)*dy1+y01;  y2 = (iy2-1)*dy2+y02

	allocate( d_ones(ncmp) );  d_ones = 1.0

	do ix2=1,nx2                               ! Zero traces have no Snell weight!
		do ix1=1,nx1
			j = grid2line3( (/nt,nx1,nx2/), (/1, ix1,ix2/) )
			k = grid2line3( (/nt,nx1,nx2/), (/nt,ix1,ix2/) )
			if( misstrace(ix1,ix2) ) d_ones(j:k) = 0.0
		end do
	end do

	snellwt = 1.0
	thresh = 10

	do jorder=0,norder
		if(jorder==0) then                       ! For m_0, don't loop over surfaces
			nnsurf = 1
		else
			nnsurf = nsurf
		end if

		do jsurf = 1,nnsurf
			iy = grid2line3( (/ny1,ny2,nsurf/), (/iy1,iy2,jsurf/) )
			itauw = grid2line3( (/nt,ny1,ny2/), (/nint((tauw(iy)-t0)/dt+1),iy1,iy2/) )
			do jsplit = 0,jorder
				call mna_set_order( jorder, jsplit, jsurf)

				k = get_cubeloc( jorder, nsurf, jsurf, jsplit)
				
				! Tabulate Snell Resampling normalization weights
				if( snellnorm) then
					stat = snell_resamp_core_exec( .true., .false., tmpsnellwt, d_ones)
					j0 = grid2line2( (/ncmp,nmod/), (/1,k/) )
					j1 = grid2line2( (/ncmp,nmod/), (/ncmp,k/) )
					snellwt(j0:j1) = tmpsnellwt
				end if

				! Tabulate reflection coefficient for each multiple generator
				do ix2=1,nx2
					x2 = (ix2-1)*dx2 + x02
					do ix1=1,nx1
						x1 = (ix1-1)*dx1 + x01
						xtot = sqrt( x1*x1 + x2*x2 )
						tmute = sqrt( ((jorder+1)*tauw(iy))**2 + (xtot/v(itauw))**2 )
						itmute = min( nt, max(1,nint( (tmute-t0)/dt - thresh )) )
						do it=itmute,nt
							tt = (it-1)*dt + t0
							jt = grid2line3( (/nt,ny1,ny2/), (/it,iy1,iy2/) )
							j = grid2line4( (/nt,nx1,nx2,nmod/), (/it,ix1,ix2,k/) )
							xp = get_xp( x1, tt, v(jt), tauw(iy), jorder, v(itauw), x2=x2 )
							refl_tab(j) = get_refl( y1, y2, x1, x2, xp, ny1, ny2, dy1, dy2,  &
							                        y01, y02, refl, jorder, jsplit, jsurf,   &
							                        nsurf, reflinv                           )
							if( it <= itmute+thresh+10 .AND. jorder > 0 ) then
								if(       reflinv) refl_tab(j) = refl_tab(j)*2.0
								if( .NOT. reflinv) refl_tab(j) = refl_tab(j)/2.0
							end if
						end do
					end do
				end do

			end do
		end do
	end do

	where( snellwt /= 0.0 ) snellwt = 1.0/snellwt 

	deallocate( d_ones)

	if(verbose) write(0,*) 'tabulate_ampl(): BOT'

end subroutine tabulate_ampl
! ------------------------------------------------------------------------------

! ------------------------------------------------------------------------------
subroutine mna_dp_test()
	real, dimension(2) :: dot1=(/0.0,0.0/), dot2=(/0.0,0.0/)

	write(0,*) '----- MULT_NMO_AMPL MODULE DOT-PRODUCT TESTS -----'
	write(0,*) '----- each row should be have same numbers -------'
	write(0,*) '----- Testing onset_mute_exec -----'
	call dot_test( onset_mute_exec, nt*nx1*nx2, nt*nx1*nx2, dot1, dot2)
	write(0,*) '----- Testing snell_resamp_exec -----'
	call dot_test( snell_resamp_exec, nt*nx1*nx2, nt*nx1*nx2, dot1, dot2)
	write(0,*) '----- Testing geometric_spreading_exec -----'
	call dot_test( geometric_spreading_exec, nt*nx1*nx2, nt*nx1*nx2, dot1, dot2)
	write(0,*) '--------------------------------------------------'

end subroutine mna_dp_test
! ------------------------------------------------------------------------------

! ------------------------------------------------------------------------------
subroutine mna_set_x02( x02_in)
	real :: x02_in
	x02 = x02_in
end subroutine mna_set_x02
! ------------------------------------------------------------------------------


end module mult_nmo_ampl_3d
