! ------------------------------------------------------------------------------
module mult_nmo_core_3d

use sep
use adj_mod
use chain0_mod
!use cartesian
use mult_nmo_util
use mult_nmo_ampl_3d
use dottest

implicit none

integer, private, save  :: nt, nx1, nx2, ny1, ny2, ordermax, korder, iy1, iy2, &
                           nmod, nsplit, isplit, ithr, ncmp, ncmps, nsurf,     &
                           isurf
real,    private, save  :: dt, t0, dx1, dx2, x01, x02, y01, y02, dy1, dy2
logical, private, save  :: verbose, test, dipop, narrow_az, snellnorm

real, dimension(:), private, pointer, save :: v, tauw, refl, dip1, dip2, vcorr
real, dimension(:), private, pointer, save :: msub, dtmp, tmp1, tmp2, tmp3
real, dimension(:), private, pointer, save :: taumult_tab, tauprim_tab,        &
                                              veff2_tab, mknown

logical, parameter, private, save :: T = .true., F = .false.

contains

! ------------------------------------------------------------------------------
subroutine mult_nmo_core_3d_init( vel, tauw_in, nt_in, t0_in, dt_in, nx1_in,   &
                                  x01_in, dx1_in, nx2_in, x02_in, dx2_in,      &
                                  ny1_in, y01_in, dy1_in, ny2_in, y02_in,      &
                                  dy2_in, maxorder, nsurf_in, refl_in,         &
                                  dip1_in, dip2_in, vcorr_in, verb, testing,   &
                                  invrefl, invgsc, ithread, dipping, narrow,   &
                                  do_gsc, do_snell, normsnell                  )

	real, intent(IN)            :: t0_in, dt_in, x01_in, dx1_in, x02_in,  &
	                               dx2_in, dy1_in, y01_in, dy2_in, y02_in
	integer, intent(IN)         :: nt_in, nx1_in, nx2_in, ny1_in, ny2_in,        &
	                               maxorder, nsurf_in
	integer                     :: ithread
	real, dimension(:), pointer :: vel, tauw_in, refl_in, dip1_in, dip2_in,      &
	                               vcorr_in
	logical                     :: verb, testing, invrefl, invgsc, dipping,      &
	                               narrow, do_gsc, do_snell, normsnell
	optional                    :: verb, testing, invrefl, invgsc, ithread,      &
	                               dipping, narrow, do_gsc, do_snell, normsnell

	verbose   = F;  if( present(verb   ) ) verbose = verb
	test      = F;  if( present(testing) ) test = testing
	ithr      = 1;  if( present(ithread) ) ithr = ithread
	dipop     = F;  if( present(dipping) ) dipop = dipping
	narrow_az = F;  if( present(narrow ) ) narrow_az = narrow
	snellnorm = F;  if( present(normsnell) ) snellnorm = normsnell
	if( .NOT. present(do_snell) ) do_snell = F
	if( .NOT. present(do_gsc)   ) do_gsc   = F

	if(verbose) write(0,*) 'mult_nmo_init(): TOP'

	nt = nt_in;     t0 = t0_in;    dt = dt_in                ! Time axis
	nx1 = nx1_in;  x01 = x01_in;  dx1 = dx1_in               ! Offset-x axis
	nx2 = nx2_in;  x02 = x02_in;  dx2 = dx2_in               ! Offset-y axis
	ny1 = ny1_in;  y01 = y01_in;  dy1 = dy1_in               ! Midpoint-x axis
	ny2 = ny2_in;  y02 = y02_in;  dy2 = dy2_in               ! Midpoint-y axis

	ordermax = maxorder
	nsurf = nsurf_in
	nmod = get_nmod(maxorder,nsurf)
	nsplit = 2**maxorder+1
	ncmp  = nt*nx1*nx2
	ncmps = ny1*ny2

	allocate( tauw(ncmps*nsurf), refl(ncmps*nsurf), v(nt*ncmps), dip1(nt*ncmps), &
	          dip2(nt*ncmps), vcorr(ncmps*nsurf)                                 )

	tauw = tauw_in;  refl = refl_in;  v = vel;  dip1 = dip1_in;  dip2 = dip2_in
	vcorr = vcorr_in

	call set_order(0,0,1)
	call mult_nmo_ampl_init( v, tauw, nt, t0, dt, nx1, x01, dx1, nx2, x02, dx2,  &
	                         ny1, y01, dy1, ny2, y02, dy2, nsurf, refl, korder,  &
	                         nmod, verb=verbose, testing=test, invrefl=invrefl,  &
	                         invgsc=invgsc, narrow=narrow_az, do_snell=do_snell, &
	                         do_gsc=do_gsc, normsnell=snellnorm                  )

	allocate( msub(ncmp), dtmp(ncmp), tmp1(ncmp), tmp2(ncmp), tmp3(ncmp),        &
	          taumult_tab(ncmp*nmod), tauprim_tab(ncmp*nmod),                    &
	          veff2_tab(ncmp*nmod), mknown(ncmp*nmod)                            )

	if(verbose) write(0,*) 'mult_nmo_init(): BOT'

end subroutine mult_nmo_core_3d_init
! ------------------------------------------------------------------------------

! ------------------------------------------------------------------------------
subroutine set_icmp( icmp1, icmp2, misstrace)
	integer :: icmp1, icmp2
	logical, dimension(:,:) :: misstrace
	optional                :: misstrace
	iy1=icmp1
	iy2=icmp2
	call tabulate_core()
	if( present(misstrace) ) then
		call mna_set_icmp(iy1,iy2,ordermax,tracemiss=misstrace)
	else
		call mna_set_icmp(iy1,iy2,ordermax)
	end if
end subroutine set_icmp
! ------------------------------------------------------------------------------

! ------------------------------------------------------------------------------
subroutine set_order(ord,spl,surf)
	integer, intent(IN) :: ord, spl, surf
	korder = ord
	isplit = spl
	isurf = surf
	call mna_set_order(korder,isplit,surf)
end subroutine set_order
! ------------------------------------------------------------------------------

! ------------------------------------------------------------------------------
subroutine mnc_set_x02( x02_in)
	real :: x02_in
	x02 = x02_in
	call mna_set_x02(x02)
end subroutine mnc_set_x02
! ------------------------------------------------------------------------------

! ------------------------------------------------------------------------------
integer function mult_nmo_exec( adj, add, model, data)
	integer             :: it0, it1, itau, jt0, jt1, jtau, ix1, ix2, iy, itauw,  &
	                       j, k, nm4(4), nd3(3), im4(4), id3(3), id4(4)
	real                :: tt, w0, w1, x1, x2, xtot2, tmax
	logical, intent(IN) :: adj, add 
	real, dimension(:)  :: model, data 

	call adjnull( adj, add, model, data)
	if(verbose) call oper_stats( 'mult_nmo_exec()', .true., model, data, adj,    &
	                             ithr=ithr                                       )

	iy = grid2line3( (/ny1,ny2,nsurf/), (/iy1,iy2,isurf/) )
	itauw = max(1, min(nt,nint( (tauw(iy)-t0)/dt+1) - 10 ) ) ! Start 10 samples up

	nm4 = (/nt,nx1,nx2,nmod/)
	nd3 = (/nt,nx1,nx2/)

	tmax = (nt-1)*dt + t0

	do ix1 = 1,nx1
		x1 = (ix1-1)*dx1 + x01
		do ix2 = 1,nx2
			x2 = (ix2-1)*dx2 + x02
			xtot2 = x1*x1 + x2*x2
			do itau = itauw,nt
				k = get_cubeloc( korder, nsurf, isurf, isplit)
				id4 = (/itau,ix1,ix2,k/);  j = grid2line4( nm4, id4 )
				tt = sqrt( (tauprim_tab(j)+taumult_tab(j))**2 + xtot2/veff2_tab(j) )

				if( tt > tmax .OR. tt < t0 ) cycle

				call get_interp1_weights( tt, dt, t0, it0, it1, w0, w1, 1, nt )
				id3 = (/it0,ix1,ix2/);  jt0  = grid2line3( nd3, id3 )
				id3 = (/it1,ix1,ix2/);  jt1  = grid2line3( nd3, id3 )
				id3 = (/itau,ix1,ix2/); jtau = grid2line3( nd3, id3 )
				if( adj) then
					model(jtau) = model(jtau) + w0*data(jt0) + w1*data(jt1)
				else
					data(jt0) = data(jt0) + w0*model(jtau)
					data(jt1) = data(jt1) + w1*model(jtau)
				end if

			end do
		end do
	end do

	if(verbose) call oper_stats( 'mult_nmo_exec()', .false., model, data, adj,   &
	                             ithr=ithr                                       )
	mult_nmo_exec = 0
end function mult_nmo_exec
! ------------------------------------------------------------------------------

! ------------------------------------------------------------------------------
integer function joint_nmo_row_exec( adj, add, model, data)
	integer                         :: iorder, im0, im1, k, stat, ksplit,     &
	                                   ksurf, nnsurf
	logical, intent(IN)             :: adj, add
	real, dimension(:)              :: model, data

	call adjnull( adj, add, model, data)
	if(verbose) call oper_stats( 'joint_nmo_row_exec()', .true., model, data,    &
	                             adj, ithr=ithr                                  )
	do iorder=0,ordermax
		if( iorder > 0 ) then                    ! For m_0, don't loop over surfaces
			nnsurf = nsurf
		else
			nnsurf = 1
		end if

		do ksurf=1,nnsurf
			do ksplit=0,iorder
				if(verbose) write(0,*) 'vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv'
				if(verbose) write(0,"(a,i1,a,i1,a,i1)") 'Ord/Surf/Spl:',iorder,',',  &
					                                       isurf,',',ksplit
				call set_order(iorder,ksplit,ksurf)

				k = get_cubeloc( iorder, nsurf, ksurf, ksplit)

				im0 = grid2line2( (/ncmp,nmod/), (/1,   k/) )
				im1 = grid2line2( (/ncmp,nmod/), (/ncmp,k/) )

				msub = 0.0;  dtmp = 0.0;  tmp1 = 0.0;  tmp2 = 0.0;  tmp3 = 0.0

				msub = model(im0:im1)
				if(adj)  dtmp = data

				call chain0( onset_mute_exec, mult_nmo_exec, snell_resamp_exec,        &
				             geometric_spreading_exec, adj, add, msub, dtmp, tmp1,     &
				             tmp2, tmp3                                                )

				if(       adj)  model(im0:im1) = msub
				if( .NOT. adj)  data = data + dtmp


				if(verbose) write(0,*) '^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^'
			end do
		end do
	end do

	joint_nmo_row_exec = 0

	if(verbose) call oper_stats( 'joint_nmo_row_exec()', .false., model, data,   &
	                             adj, ithr=ithr                                  )
end function joint_nmo_row_exec
! ------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
! For each CMP, tabulate values used for HEMNO operator.
!--------------------------------------------------------------------------------
subroutine tabulate_core()
	integer :: itau, ix1, ix2, iy, itauw, jy, j, jsurf, nnsurf, jorder, k,        &
	           jsplit, nd3(3), nm4(4), ny3(3), my3(3), id3(3), im4(4), ky
	real    :: tau, x1, x2, y1, y2, xp, ypr(2), vprim, vmult, vcos

	if(verbose) write(0,*) 'tabulate_core() - TOP'

	y1 = (iy1-1)*dy1 + y01
	y2 = (iy2-1)*dy2 + y02

	ny3 = (/ny1,ny2,nsurf/)	
	my3 = (/nt,ny1,ny2/)	
	nm4 = (/nt,nx1,nx2,nmod/)

	do jorder = 0,ordermax
		if( jorder > 0 ) then                    ! For m_0, don't loop over surfaces
			nnsurf = nsurf
		else
			nnsurf = 1
		end if

		do jsurf = 1,nnsurf
			id3 = (/iy1,iy2,jsurf/);  iy = grid2line3( ny3, id3 )
			id3 = (/nint((tauw(iy)-t0)/dt+1),iy1,iy2/)
			itauw = grid2line3( my3, id3 )
			ky = grid2line3( (/ny1,ny2,nsurf/), (/iy1,iy2,jsurf/) )
			do jsplit = 0,jorder

				k = get_cubeloc( jorder, nsurf, jsurf, jsplit)

				do ix1 = 1,nx1
					x1 = (ix1-1)*dx1 + x01
					do ix2 = 1,nx2
						x2 = (ix2-1)*dx2 + x02
						do itau = 1,nt
							tau = t0 + (itau-1)*dt
							id3 = (/itau,iy1,iy2  /);  jy = grid2line3( my3, id3 )
							im4 = (/itau,ix1,ix2,k/);   j = grid2line4( nm4, im4 )
							xp = get_xp( x1, tau, v(jy), tauw(iy), jorder, v(itauw), x2=x2 )
							if( dipop) then                                      ! Dipping NMO
								taumult_tab(j) = get_taumult( y1, y2, x1, x2, xp, ny1, ny2,    &
								                              dy1, dy2, y01, y02, tauw,        &
								                              jorder, jsplit, jsurf, nsurf     )
								call get_prim_yrefl( y1, y2, xp, x1, x2, ypr, jsplit, jorder)
								tauprim_tab(j) = get_tau3( y1, y2, ypr(1), ypr(2), tau, ny1,   &
								                           dy1, y01, ny2, dy2, y02, nt, dt,    &
								                           t0, dip1, dip2, jorder, jsplit      )
							else                                                     ! 1-D NMO
								taumult_tab(j) = jorder*tauw(iy)
								tauprim_tab(j) = tau
							end if

							veff2_tab(j) = get_veff2( y1, y2, x1, x2, xp, tauprim_tab(j),    &
							                          ny1, ny2, dy1, dy2, y01, y02, nt, dt,  &
							                          t0, v, tauw, jorder, jsplit, jsurf,    &
							                          nsurf                                  )

							if(jorder>0) veff2_tab(j) = veff2_tab(j)*vcorr(ky)
						end do
					end do
				end do
			end do
		end do
	end do

	if(verbose) write(0,*) 'tabulate_core() - BOT'

end subroutine tabulate_core
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
subroutine mult_nmo_core_close( )
	if(verbose) write(0,*) 'mult_nmo_core_close(): TOP'
	if( test) then
		if( .NOT. allocated(v)    )  call seperr( 'ERROR! Array v not allocated'   )
		if( .NOT. allocated(tauw) )  call seperr( 'ERROR! Array tauw not allocated')
		if( .NOT. allocated(refl) )  call seperr( 'ERROR! Array refl not allocated')
		if( .NOT. allocated(dip1) )  call seperr( 'ERROR! Array dip1 not allocated')
		if( .NOT. allocated(dip2) )  call seperr( 'ERROR! Array dip2 not allocated')
		if( .NOT. allocated(msub) )  call seperr( 'ERROR! Array msub not allocated')
		if( .NOT. allocated(dtmp) )  call seperr( 'ERROR! Array dtmp not allocated')
		if( .NOT. allocated(tmp1) )  call seperr( 'ERROR! Array tmp1 not allocated')
		if( .NOT. allocated(tmp2) )  call seperr( 'ERROR! Array tmp2 not allocated')
		if( .NOT. allocated(tmp3) )  call seperr( 'ERROR! Array tmp3 not allocated')
		if( .NOT. allocated(vcorr))  call seperr( 'ERROR! Array vcorr not allocated')
		if( .NOT. allocated(taumult_tab) ) &
			call seperr( 'ERROR! Array taumult_tab not allocated')
		if( .NOT. allocated(tauprim_tab) ) & 
			call seperr( 'ERROR! Array tauprim_tab not allocated')
		if( .NOT. allocated(veff2_tab)    ) & 
			call seperr( 'ERROR! Array veff2_tab not allocated')
		if( .NOT. allocated(mknown)    ) & 
			call seperr( 'ERROR! Array mknown not allocated')
	end if
	deallocate( v, tauw, refl, dip1, dip2, msub, dtmp, tmp1, tmp2, tmp3,         &
	            taumult_tab, tauprim_tab, veff2_tab, mknown                       )
	call mult_nmo_ampl_close()
	if(verbose) write(0,*) 'mult_nmo_core_close(): BOT'
end subroutine mult_nmo_core_close
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
subroutine mnc_dp_test()
	real :: dot1(2), dot2(2)

	write(0,*) '----- MULT_NMO_CORE MODULE DOT-PRODUCT TESTS -----'
	write(0,*) '----- each row should be have same numbers -------'
	write(0,*) '----- Testing mult_nmo_exec -----'
	dot1=0.0;  dot2=0.0
	call dot_test( mult_nmo_exec, ncmp, ncmp, dot1, dot2)
	write(0,*) '----- Testing joint_nmo_row_exec -----'
	call dot_test( joint_nmo_row_exec, nmod*ncmp, ncmp, dot1, dot2)
	write(0,*) '--------------------------------------------------'

end subroutine mnc_dp_test
!--------------------------------------------------------------------------------
   
!--------------------------------------------------------------------------------
subroutine mult_nmo_core_3d_test()
	real, dimension(:), pointer :: v, tw, r, p1, p2, vc

	allocate( v(100*5), tw(5*2), r(5*2), p1(100*5), p2(100*5), vc(5*2) )

	v = 1500.0
	tw(1:5)  = 0.1
	tw(6:10) = 0.2
	r(1:5)  = -0.2
	r(6:10) = -0.1
	p1 = 0.0
	p2 = 0.0
	vc = 1.0
	
	call mult_nmo_core_3d_init( v, tw, 100, 0.0, 0.004, 20, 0.0, 25.0, 1, 0.0,    &
	                            25.0, 5, 0.0, 50.0, 1, 0.0, 50.0, 1, 2, r, p1,    &
	                            p2, vc                                            )
	call set_icmp( 3, 1)

	call mnc_dp_test()

	call mult_nmo_core_close()

end subroutine mult_nmo_core_3d_test
!--------------------------------------------------------------------------------

end module mult_nmo_core_3d
