! ----------------------------------------
!  picks2seplib.f90 - Created by Morgan Brown
!           Tue Feb 20 10:29:46 2001.
! ----------------------------------------
!       Usage: picks2seplib.x fname=filename.txt [order=1,2,3] > tv.H
!  Parameters: fname - Name of Rickmovie pick file (less than 50 characters).
!              order - integer(3) - Default = 1,2,3, which writes out data in
!                                   t, x, y order.  If you wanted x as the fast
!                                   axis and t as the second axis, order = 2,1,3.
! Description: Takes a Rickmovie pick file and writes the abcissas and ordinates to
!              a sepfile.
! ----------------------------------------
program picks2seplib

use sep

implicit none

integer                     :: n1, n2, fstat, ipick, npick, order(3)
real                        :: d1, d2, o1, o2
real, dimension(:), pointer :: t, x, y
character(50)               :: fname

call sep_init(SOURCE)
write(0,*) 'after sep_init()'
call from_param('fname',fname)

order = (/1,2,3/)
call from_param('order',order,order)

write(0,*) 'after from_param()'

2000 format( 2x, f10.0, 1x, f10.0, 1x, f10.0, 33x)

  open( unit=20, file=fname, status='old')
	npick = 0; fstat=0
  do while( fstat==0 )                       ! Determine length of pickfile
    read( 20, 2000, iostat=fstat )
		npick = npick + 1
  end do
  close( unit=20)

  open( unit=20, file=fname, status='old')
	allocate( t(npick-1), x(npick-1), y(npick-1) )
	do ipick=1,npick-1
    read( 20, 2000, iostat=fstat )  t(ipick), x(ipick), y(ipick)
	end do
  close( unit=20)
	
	call to_history( 'n1', npick-1)
	call to_history( 'n2', 3)
	call to_history( 'esize', 4)

	call sep_close()

	if( order(1)==1) then
		call sep_write(t)
		if( order(2)==2) then  ! Order = 1,2,3
			call sep_write(x)
			call sep_write(y)
		else                   ! Order = 1,3,2
			call sep_write(y)
			call sep_write(x)
		end if
	else if( order(1)==2 ) then
		call sep_write(x)
		if( order(2)==1) then  ! Order = 2,1,3
			call sep_write(t)
			call sep_write(y)
		else                   ! Order = 2,3,1
			call sep_write(y)
			call sep_write(t)
		end if
	else if( order(1)==3 ) then
		call sep_write(y)
		if( order(2)==1) then  ! Order = 3,1,2
			call sep_write(t)
			call sep_write(x)
		else                   ! Order = 3,2,1
			call sep_write(x)
			call sep_write(t)
		end if
	end if

	call exit(0)

end program picks2seplib
