!-------------------------------------------------------------------------------
!  polyfit.f90 - Created by Morgan Brown
!           Mon Sep  8 22:02:41 2003.
!-------------------------------------------------------------------------------
!       Usage: polyfit.mod 
!  Parameters: 
! Description: 
!-------------------------------------------------------------------------------
module polyfit

use sep
use adj_mod
use dottest
use morgan_util
use Cubic_Spline_GCV
use qshep2d
!use qshep3d

implicit none

logical, private, save :: verbose, tz_func
logical, parameter, private, save :: T=.true., F=.false.
real, dimension(:), pointer, private, save :: tmp

contains

!-------------------------------------------------------------------------------
subroutine polyfit_init( verb )
	logical  :: verb
	optional :: verb

	verbose=F;  if(present(verb) ) verbose=verb
end subroutine polyfit_init
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
subroutine polyfit_close( )
end subroutine polyfit_close
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
integer function polyfit_exec( adj, add, model, data)
	logical            :: adj, add 
	real, dimension(:) :: model, data 

	call adjnull( adj, add, model, data)

	if( verbose) call oper_stats( 'polyfit_exec', T, model, data, adj)

	if( adj) then

	else

	end if

	if( verbose) call oper_stats( 'polyfit_exec', F, model, data, adj)

	polyfit_exec = 0

end function polyfit_exec
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
subroutine polyfit_test( )
	real    :: dot1(2), dot2(2)
	integer :: nmod, ndat

	call polyfit_init( )
	call dot_test( polyfit_exec, nmod, ndat, dot1, dot2)
end subroutine polyfit_test
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
subroutine quadratic_spline_surf( txy, surf, ny1, ny2, dy1, dy2, y01, y02 )
	integer :: nd, nq, nw, nr, ier, ny1, ny2, iy1, iy2, k
	real    :: dy1, dy2, y01, y02, y1, y2
	real    :: xmin, ymin, dx, dy, rmax, y1k, y2k, fx, fy

	real,    dimension(:,:)              :: txy, surf
	real,    dimension(:),   allocatable :: rsq
	real,    dimension(:,:), allocatable :: a
	integer, dimension(:),   allocatable :: lnext
	integer, dimension(:,:), allocatable :: lcell

	nd = size(txy,dim=1)

	nq = 13
	nw = 19
	nr = nint( sqrt(nd/3.0) )

	allocate( lcell(nr,nr), lnext(nd), rsq(nd), a(5,nd) )
	
	write(0,*) 'quadratic_spline_surf(): bytes allocated=',nr*nr+7*nd

	call qshep2( nd, txy(:,2), txy(:,3), txy(:,1), nq, nw, nr, lcell, lnext,     &
	             xmin, ymin, dx, dy, rmax, rsq, a, ier                           )

	do iy2=1,ny2
		y2 = (iy2-1)*dy2 + y02
		do iy1=1,ny1
			y1 = (iy1-1)*dy1 + y01
			call qs2grd( y1, y2, nd, txy(:,2), txy(:,3), txy(:,1), nr, lcell, lnext, &
			             xmin, ymin, dx, dy, rmax, rsq, a, surf(iy1,iy2), fx, fy, ier)
		end do
	end do

	deallocate( lcell, lnext, rsq, a )

end subroutine quadratic_spline_surf
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
subroutine quadratic_spline_dipsurf( txy, dipsurf, ny1, ny2, dy1, dy2, y01, y02 )
	integer :: nd, nq, nw, nr, ier, ny1, ny2, iy1, iy2, k
	real    :: dy1, dy2, y01, y02, y1, y2
	real    :: xmin, ymin, dx, dy, rmax, y1k, y2k, f

	real,    dimension(:,:)              :: txy
	real,    dimension(:,:,:)            :: dipsurf
	real,    dimension(:),   allocatable :: rsq
	real,    dimension(:,:), allocatable :: a
	integer, dimension(:),   allocatable :: lnext
	integer, dimension(:,:), allocatable :: lcell

	nd = size(txy,dim=1)

	nq = 13
	nw = 19
	nr = nint( sqrt(nd/3.0) )

	allocate( lcell(nr,nr), lnext(nd), rsq(nd), a(5,nd) )
	
	write(0,*) 'quadratic_spline_surf(): bytes allocated=',nr*nr+7*nd

	call qshep2( nd, txy(:,2), txy(:,3), txy(:,1), nq, nw, nr, lcell, lnext,     &
	             xmin, ymin, dx, dy, rmax, rsq, a, ier                           )

	do iy2=1,ny2
		y2 = (iy2-1)*dy2 + y02
		do iy1=1,ny1
			y1 = (iy1-1)*dy1 + y01
			call qs2grd( y1, y2, nd, txy(:,2), txy(:,3), txy(:,1), nr, lcell, lnext, &
			             xmin, ymin, dx, dy, rmax, rsq, a, f, dipsurf(iy1,iy2,1),    &
			             dipsurf(iy1,iy2,2), ier                                     )
		end do
	end do

	deallocate( lcell, lnext, rsq, a )

end subroutine quadratic_spline_dipsurf
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
subroutine cubic_spline( xn4, fn4, nx, dx4, x04, fx, dip, variance)
	real,    intent(IN) :: dx4, x04
	real (dp)           :: dx, x0
	integer, intent(IN) :: nx
	logical, intent(IN) :: dip
	real, dimension(:), intent(IN)  :: xn4, fn4
	real, dimension(:), intent(OUT) :: fx
	real (dp), dimension(:), allocatable :: xn, fn, err, y, se, tmp
	real (dp), dimension(:,:), allocatable :: c
	integer :: nd, job, ierr, ix, ispl
	real (dp)    :: var, d, x
	real :: variance
	optional :: variance

	var=-1.0;  if(present(variance)) var=variance

	dx = dx4                                                    ! type conversion
	x0 = x04                                                    ! type conversion

	job = 1
	nd = size(xn4,dim=1)

	allocate( xn(nd), fn(nd), err(nd), y(nd), c(nd-1,3), tmp(7*(nd+2)), se(nd) )

	xn = xn4                                                    ! type conversion
	fn = fn4                                                    ! type conversion

	err = 1.0

	call cubgcv( xn, fn, err, nd, y, c, nd-1, var, job, se, tmp, ierr)

	ispl = 1
	do ix=1,nx
		x = (ix-1)*dx + x0
		if( x < xn(1) ) then
			if( dip) then
				fx(ix) = 0.0
			else
				fx(ix) = fn(ispl)
			end if
		else if( x >= xn(ispl) ) then
			if( ispl < nd-1) then
				if( x >= xn(ispl+1) ) ispl = ispl + 1
				d = x-xn(ispl)
				if( dip) then
					fx(ix) = 3*c(ispl,3)*d*d + 2*c(ispl,2)*d + c(ispl,1)
				else
					fx(ix) = c(ispl,3)*d*d*d + c(ispl,2)*d*d + c(ispl,1)*d + y(ispl)
				end if
			else
				if( dip) then
					fx(ix) = 0.0
				else
					fx(ix) = fn(nd)
				end if
			end if
		end if
	end do	

	deallocate( xn, fn, err, y, c, tmp, se )

end subroutine cubic_spline
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! After interpolating axis dim, count number of new data points
!-------------------------------------------------------------------------------
! Example: 200 velocity picks in 24 functions might have 10000 data points after
!          gridding the t-axis
!-------------------------------------------------------------------------------
integer function count_gridded( xyz, dim, nd, d1, d2, d3, n1, n2, n3, extrap)  &
                                                               result( nouttot )
  integer,                 intent(IN) :: nd, dim, n1, n2, n3
	real,                    intent(IN) :: d1, d2, d3
  real,    dimension(:,:), intent(IN) :: xyz
	logical,                 intent(IN) :: extrap
  integer                             :: ndtot, ct, j0, j1, ngtmp, n(3)
	real                                :: d(3)

	ndtot = 0
	nouttot = 0
	d = (/d1,d2,d3/)
	n = (/n1,n2,n3/)

	do while( ndtot+1 < nd) 
		ct = get_func_count( nd, ndtot, dim, xyz, d1, d2, d3)

		j0 = ndtot+1
		j1 = j0+ct-1
		ndtot = ndtot + ct
		if(j1>nd) cycle

		if( extrap ) then
			ngtmp = n(dim)
		else
			ngtmp = nint( (xyz(j1,dim)-xyz(j0,dim))/d(dim) + 1)
		end if

		if( ct > 4)  nouttot = nouttot + ngtmp        ! Ignore short (ngtmp<4) funcs
	end do

end function count_gridded
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! Count irregular data samples in a 1-D function
!-------------------------------------------------------------------------------
integer function get_func_count( nd, ndtot, dim, xyz, d1, d2, d3) result(ct)
  integer,                 intent(IN) :: nd, ndtot, dim
  real,    dimension(:,:), intent(IN) :: xyz
  integer                             :: j0, j1
	real                                :: d1, d2, d3, dist1, dist2, dm1, dm2
	logical                             :: done

	ct = 1
	done = F
	do while( .NOT. done)
		j0 = ndtot+ct
		if(j0==nd) then
			done=T
			cycle
		end if
		j1 = j0+1

		if( dim==1 ) then
			dist1 = xyz(j1,2) - xyz(j0,2)
			dist2 = xyz(j1,3) - xyz(j0,3)
			dm1 = d2;  dm2 = d3
		else if( dim==2) then
			dist1 = xyz(j1,1) - xyz(j0,1)
			dist2 = xyz(j1,3) - xyz(j0,3)
			dm1 = d1;  dm2 = d3
		else if( dim==3) then
			dist1 = xyz(j1,1) - xyz(j0,1)
			dist2 = xyz(j1,2) - xyz(j0,2)
			dm1 = d1;  dm2 = d2
		else
			call seperr('get_func_count(): dim=1,2, or 3')
		end if

		if( abs(dist1) >= dm1/2.0 .OR. abs(dist2) >= dm2/2.0 ) then
			done = T
			cycle
		end if
		ct = ct + 1
	end do

end function get_func_count
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! After interpolating one axis, we need to reorder the irregular data such that
! the fast axis is along the next direction of interpolation.
!-------------------------------------------------------------------------------
! Example: Interpolate velocity functions along 1-axis.  Reorder along 2-axis,
!          then interpolate along 2-axis
!-------------------------------------------------------------------------------
subroutine reorder_funcs( nd, dim, gridval, xyz, val, d1, d2, d3, o1, o2, o3,  &
                          n1, n2, n3                                           )

  integer,                 intent(IN)  :: nd, dim, n1, n2, n3
  real,                    intent(IN)  :: d1, d2, d3, o1, o2, o3
  real,    dimension(:),   intent(IN)  :: gridval
  real,    dimension(:),   intent(OUT) :: val
  real,    dimension(:,:), intent(OUT) :: xyz
  integer                              :: nm1, nm2, nm3, im1, im2, im3, ct,    &
	                                        id, j
	real                                 :: dm1, dm2, dm3, om1, om2, om3, xc1,   &
	                                        xc2, xc3

	if( dim==1 ) then                                          ! Fast->slow: 1,2,3
		nm1 = n1;  nm2 = n2;  nm3 = n3
		dm1 = d1;  dm2 = d2;  dm3 = d3
		om1 = o1;  om2 = o2;  om3 = o3
	else if( dim==2) then                                      ! Fast->slow: 2,1,3
		nm1 = n2;  nm2 = n1;  nm3 = n3
		dm1 = d2;  dm2 = d1;  dm3 = d3
		om1 = o2;  om2 = o1;  om3 = o3
	else if( dim==3) then                                      ! Fast->slow: 3,1,2
		nm1 = n3;  nm2 = n1;  nm3 = n2
		dm1 = d3;  dm2 = d1;  dm3 = d2
		om1 = o3;  om2 = o1;  om3 = o2
	else
		call seperr('get_func_count(): must have dim=1,2, or 3')
	end if

	ct = 1

	do im3=1,nm3
		xc3 = (im3-1)*dm3 + om3
		do im2=1,nm2
			xc2 = (im2-1)*dm2 + om2
			do im1=1,nm1
				xc1 = (im1-1)*dm1 + om1

				if( dim==1 ) then
					j = grid2line3( (/n1,n2,n3/), (/im1,im2,im3/) )
				else if( dim==2) then
					j = grid2line3( (/n1,n2,n3/), (/im2,im1,im3/) )
				else if( dim==3) then
					j = grid2line3( (/n1,n2,n3/), (/im2,im3,im1/) )
				else
					call seperr('get_func_count(): dim=1,2, or 3')
				end if

				if( gridval(j) /= 0.0 ) then                        ! Don't use fold directly
					if( dim==1 ) then
						xyz(ct,:) = (/xc1,xc2,xc3/)
					else if( dim==2) then
						xyz(ct,:) = (/xc2,xc1,xc3/)
					else if( dim==3) then
						xyz(ct,:) = (/xc2,xc3,xc1/)
					else
						call seperr('get_func_count(): dim=1,2, or 3')
					end if
!					if( ct > nd) cycle
					val(ct) = gridval(j)
					ct = ct + 1
				end if

			end do
		end do
	end do

end subroutine reorder_funcs
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! One pass of cubic spline interpolation of a 1-D function
!-------------------------------------------------------------------------------
subroutine picks_to_gridded_func( xyz, val, nd, dim, d1, d2, d3, o1, o2, o3,   &
                                  n1, n2, n3, xyzout, valout   )

	integer,                 intent(IN)  :: nd, dim, n1, n2, n3
	real,                    intent(IN)  :: d1, d2, d3, o1, o2, o3
  real,    dimension(:),   intent(IN)  :: val
  real,    dimension(:,:), intent(IN)  :: xyz
  real,    dimension(:),   intent(OUT) :: valout
  real,    dimension(:,:), intent(OUT) :: xyzout
  integer :: ndtot, ct, j0, j1, ngtmp, ig, k0, k1, n(3), ik
  real    :: d(3), o(3)

	k0=0
	k1=0
	ndtot = 0
	d = (/d1,d2,d3/)
	o = (/o1,o2,o3/)
	n = (/n1,n2,n3/)

	do while( ndtot+1 < nd) 
		ct = get_func_count( nd, ndtot, dim, xyz, d1, d2, d3)
		j0 = ndtot + 1
		j1 = j0 + ct - 1
		ndtot = ndtot + ct
		if(j1>nd) cycle

		k0 = k1 + 1
		k1 = k0 + n(dim) - 1
		if( k1 > size(valout) ) cycle
		if(ct<=4) cycle

		call cubic_spline( xyz(j0:j1,dim), val(j0:j1), n(dim), d(dim), o(dim),   &
		                   valout(k0:k1), F                                      )
		do ig=1,n(dim)
			ik = k0 + ig - 1
			xyzout(ik,dim) = (ig-1)*d(dim) + o(dim)                        ! gridded
			if( dim==1 ) then
				xyzout(ik,2) = xyz(j0,2)
				xyzout(ik,3) = xyz(j0,3)
			else if( dim==2 ) then
				xyzout(ik,1) = xyz(j0,1)
				xyzout(ik,3) = xyz(j0,3)
			else if( dim==3 ) then
				xyzout(ik,1) = xyz(j0,1)
				xyzout(ik,2) = xyz(j0,2)
			else
				call seperr('picks_to_gridded_func(): dim must=1,2, or 3')
			end if
		end do

	end do

end subroutine picks_to_gridded_func
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
subroutine interpolate_surfs_in_time( txysurf, psurf, n1, n2, n3, d1, d2, d3,  &
                                      o1, o2, o3, pgrid                        )
	integer, intent(IN) :: n3, n2, n1
	real,    intent(IN) :: d3, d2, d1, o3, o2, o1
	integer             :: i3, i2, i1, ct, ndsurf, id, i1d, i2d, i3d, imin1,     &
	                       imax1
	real, dimension(:)   :: psurf
	real, dimension(:,:) :: txysurf
	real, dimension(:,:,:) :: pgrid
	real, dimension(:), allocatable :: ttmp, ftmp

	ndsurf = size(txysurf,dim=1)
	pgrid = 0.0

	do i3=1,n3
		do i2=1,n2
			write(0,"(a,i5,a,i5,a,i5,a,i5)") 'done midpoint',i2,' of',n2,            &
			                                       ', slice',i3,' of',n3
			ct = 0
			do i1=1,n1
				do id=1,ndsurf
					i1d = nint( (txysurf(id,1)-o1)/d1 + 1);  if( i1d /= i1) cycle
					i2d = nint( (txysurf(id,2)-o2)/d2 + 1);  if( i2d /= i2) cycle
					i3d = nint( (txysurf(id,3)-o3)/d3 + 1);  if( i3d /= i3) cycle
					ct = ct + 1	
				end do
			end do

			if( ct < 5 ) cycle

			allocate( ttmp(ct+2), ftmp(ct+2) )

			ttmp(1) = o1
			ttmp(ct+2) = (n1-1)*d1 + o1
			ftmp = 0.0

			ct = 1
			do i1=1,n1
				do id=1,ndsurf
					i1d = nint( (txysurf(id,1)-o1)/d1 + 1);  if( i1d /= i1) cycle
					i2d = nint( (txysurf(id,2)-o2)/d2 + 1);  if( i2d /= i2) cycle
					i3d = nint( (txysurf(id,3)-o3)/d3 + 1);  if( i3d /= i3) cycle
					ct = ct + 1	
					ttmp(ct) = txysurf(id,1)
					ftmp(ct) = psurf(id)
				end do
			end do

!			imin1 = nint( (ttmp(1) -o1)/d1 + 1)
!			imax1 = nint( (ttmp(ct)-o1)/d1 + 1)

			call cubic_spline( ttmp, ftmp, n1, d1, o1, pgrid(:,i2,i3), F, variance=-1.0 )
!			call cubic_spline( ttmp, ftmp, imax1-imin1, d1, ttmp(1),                 &
!			                   pgrid(imin1:imax1,i2,i3), F, variance=-1.0            )

			deallocate( ttmp, ftmp)

		end do
	end do

end subroutine interpolate_surfs_in_time
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
subroutine count_interpolated_samples( txy, nd, d1, d2, nouttot )
  integer, intent(IN)  :: nd
	real,    intent(IN)  :: d1, d2
  integer, intent(OUT) :: nouttot
  real, dimension(:,:), intent(IN) :: txy
  integer :: ndtot, ct, j0, j1, nxtmp
	real    :: dist, davg, tdist, xdist
	logical :: done

	ndtot = 0
	nouttot = 0
	davg = 0.0
	do while( ndtot+1 < nd) 
		ct = get_surf_count( nd, ndtot, txy, d1, d2)

		j0 = ndtot+1
		j1 = j0+ct-1
		ndtot = ndtot + ct
		if(j1>nd) cycle

		nxtmp = nint( (txy(j1,2)-txy(j0,2))/d2 + 1)

		if( ct > 4) nouttot = nouttot+nxtmp
	end do

end subroutine count_interpolated_samples
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
integer function get_surf_count( nd, ndtot, txy, d1, d2) result(ct)
  integer, intent(IN) :: nd, ndtot
  integer :: j0, j1, nxtmp, ix
	real    :: dist, d1, d2, davg, tdist, xdist
  real, dimension (:,:) :: txy
	logical :: done

	ct = 1
	done = F
	do while( .NOT. done)
		j0 = ndtot+ct
		if(j0==nd) then
			done=T
			cycle
		end if
		j1 = j0+1

		xdist = (txy(j1,2) - txy(j0,2))/d2             ! abcissae must be ordered
		tdist = (txy(j1,1) - txy(j0,1))/d1
		dist = sqrt( tdist**2 + xdist**2 )             ! in samples
		if( xdist <= 0.0 .OR. dist > 20 .OR. txy(j1,3)/=txy(j0,3) ) then
			done = T
			cycle
		end if
		ct = ct + 1
	end do

end function get_surf_count
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
subroutine picks_to_gridded_dipsurf( txy, nd, d1, d2, pout, txyout )

  integer :: nd, ndtot, ct, j0, j1, nxtmp, ix, k0, k1
	real    :: d1, d2, max1, min1
  real, dimension (:), pointer :: xtmp, ftmp, pout
  real, dimension (:,:), pointer :: txy, txyout

	k0=0
	k1=0
	ndtot = 0
	do while( ndtot+1 < nd) 

		ct = get_surf_count( nd, ndtot, txy, d1, d2)

		j0 = ndtot+1
		j1 = j0+ct
		ndtot = ndtot + ct
		if(j1>nd) cycle
		max1 = maxval(txy(j0:j1,1))
		min1 = minval(txy(j0:j1,1))

		allocate( xtmp(ct), ftmp(ct) )

		xtmp = txy(j0:j1,2)
		ftmp = txy(j0:j1,1)

		nxtmp = nint( (xtmp(ct)-xtmp(1))/d2 + 1)

		if( ct > 4) then
			k0 = k1 + 1
			k1 = k0 + nxtmp
			call cubic_spline( xtmp, ftmp, nxtmp, d2, xtmp(1), txyout(k0:k1,1), F,    &
			                   variance=-1.0 )
			
			where( txyout(k0:k1,1) > max1) txyout(k0:k1,1)=max1
			where( txyout(k0:k1,1) < min1) txyout(k0:k1,1)=min1

			do ix=1,nxtmp
				txyout(k0+ix,2) = (ix-1)*d2 + xtmp(1)
				txyout(k0+ix,3) = txy(j0,3)
			end do

			call cubic_spline( xtmp, ftmp, nxtmp, d2, xtmp(1), pout(k0:k1), T)

		end if

		deallocate( xtmp, ftmp)

	end do

end subroutine picks_to_gridded_dipsurf
!-------------------------------------------------------------------------------

end module polyfit
