! Findwarp
!
! Local cross-correlation on a brick-by-brick basis
! Finds shifts in x, y and z
! Units are samples
! Warping itself done elsewhere (Warp.x)
!
! USAGE
!
! Findwarp.x <slave.H master=master.H shift1=shift1.H shift2=shift2.H ...
!
! PARAMETERS
!
!		maxshift1,maxshift2,maxshift3	- [int] maximum shifts
!		nb1=n1, nb2=1, nb3=1	- [int] brick size
!		nm1=n1, nm2=n2, nm3=60	- [int] size of chunk held in-core
!		j1, j2, j3 		- [int] subsample 
!		wantxcorr=0		- dump x-correlation functions to
!						xcorr=xcorr.H (2D only)
!
! (C) James Rickett - 1997
!
!%
program findwarp
	use tdxcorr
  use sep
  implicit none

  integer n1,n2,n3,n4,n5, nb1,nb2,nb3, j1,j2,j3, nm1,nm2,nm3
  real, allocatable, dimension(:,:) :: brick1,brick2
  real, allocatable, dimension(:)   :: xcorr, shift
  real :: o1,o2,o3,o4,o5,d1,d2,d3,d4,d5
  integer :: ishift1,ishift2,ishift3
  integer :: n_shift1,n_shift2,n_shift3
  integer :: fm11,fm12,fm13, fm21,fm22,fm23, sf1,sf2	! brickbybrick stuff
  integer :: i1,i2,i3,i4,i5             ! counters
  integer :: maxshift1,maxshift2,maxshift3,wantxcorr

  call sep_init(SOURCE)

	call from_history('n1',n1)
	call from_history('n2',n2)
	call from_history('n3',n3)
	call from_history('n4',n4)
	call from_history('n5',n5)

  allocate( brick1(n1,n2), brick2(n1,n2) )

  allocate( shift(n1), xcorr(2*n1-1) )

  call from_param('maxshift1',maxshift1,4)

  call from_history('o1',o1,0.)
  call from_history('o2',o2,0.)
  call from_history('o3',o3,0.)
  call from_history('o4',o4,0.)
  call from_history('o5',o5,0.)
  call from_history('d1',d1,1.)
  call from_history('d2',d2,1.)
  call from_history('d3',d3,1.)
  call from_history('d4',d4,1.)
  call from_history('d5',d5,1.)

  call to_history('n1',n1,'shift1')
  call to_history('n2',n2,'shift1')
  call to_history('n3',n3,'shift1')
  call to_history('n4',n4,'shift1')
  call to_history('n5',n5,'shift1')
  call to_history('o1',o1,'shift1')
  call to_history('o2',o2,'shift1')
  call to_history('o3',o3,'shift1')
  call to_history('o4',o4,'shift1')
  call to_history('o5',o5,'shift1')
  call to_history('d1',d1,'shift1')
  call to_history('d2',d2,'shift1')
  call to_history('d3',d3,'shift1')
  call to_history('d4',d4,'shift1')
  call to_history('d5',d5,'shift1')

  call sep_close()

!-------------------------------------------------------------------------------
	do i5=1,n5
		do i4=1,n4
			do i3=1,n3
				call sep_read( brick1, 'master')
				call sep_read( brick2 )
				do i2=1,n2
					do i1=1,n1
						call tdxcorr( n1, brick1(:,i2), brick2(:,i2), xcorr, shift(i1), maxshift1 )
					end do
					call srite('shift1',shift(1),n1*4)
				end do
			end do
		end do
	end do
!-------------------------------------------------------------------------------


end program findwarp

