! N-dimensional non-stationary match filter program
program MatchND
  use helix
  use nhelix
  use solver_mod
  use cgstep_mod
  use sep 
  use tohist
  use ncnhconest
  use nchelicon
  use ncnhelicon
  use polydiv
  use polydiv2
  use matchprec
  use helicon
  use pef
  use cartesian

  implicit none
  real,    dimension (:), pointer   :: data1, data2, res
  real,    dimension (:), pointer   :: aa,bb

  integer, dimension(:), pointer :: nf,psize,defaults,np,maxshift,ii
  integer :: niter,ndim,nn,nnp,fdim,nfilt,j,k,index
  real    :: eps,e1,e2

  integer stat,i,nhmax
  type (nfilter) :: ff
  type (filter)  :: laplac,pef1

  integer, allocatable, dimension(:) :: nh,pch
  integer, pointer, dimension(:) :: n,naux,nout
  real,    pointer, dimension(:) :: o,d,dout,oout,err
  real,    pointer, dimension(:,:) :: xmov

  logical, parameter :: wantmatched=.false.
  logical :: verb

  call sep_init (SOURCE)
  ndim=sep_dimension()
  fdim=ndim

  verb=.false.
  if (fdim==2) verb=.true.
  allocate(nf(ndim),psize(ndim),defaults(ndim),ii(ndim),naux(ndim))
  allocate(np(ndim))
  defaults=1
  call from_param('nf',nf,defaults)
  call from_param('patchsize',psize,defaults)
  call from_param('niter',niter,product(nf))
  call from_param('eps',eps,1.0e4)

  allocate(n(ndim),d(ndim),o(ndim),err(niter))
  call from_history ('n',n)
  call from_history ('d',d)
  call from_history ('o',o)

  call from_aux('master','n',naux)
  nn=1
  do i=1,ndim
     if(n(i).ne.naux(i)) call seperr('n doesnt match')
     nn=nn*n(i)
  end do

  ! calculate number of patches
  allocate (np(ndim))
  do i=1,ndim
     np(i)=1+(n(i)-1)/(psize(i))
  end do
  nnp=product(np)

  ! write headers for filt
  allocate(nout(ndim+fdim+1),dout(ndim+fdim+1),oout(ndim+fdim+1),maxshift(fdim))
  nout(     1:     fdim)=nf(1:fdim)
  nout(1+fdim:ndim+fdim)=np(1:ndim)

  do i=1,fdim
     nout(i)=nf(i)
     dout(i)= d(i)
     oout(i)=-d(i)*((nf(i)-1)/2)
     maxshift(i)=(nf(i)-1)/2
  end do
  do i=1,ndim
     nout(i+fdim)=np(i)
     dout(i+fdim)=d(i)*psize(i)
     oout(i+fdim)=o(i)
  end do
!  nout(ndim+fdim+1)=niter
  nout(ndim+fdim+1)=1
  dout(ndim+fdim+1)=1.
  oout(ndim+fdim+1)=0.

  call to_hist_int_array('n',nout,'filt')
  call to_hist_float_array('d',dout,'filt')
  call to_hist_float_array('o',oout,'filt')
  call to_history('esize',4,'filt')

  call to_history('n1',niter,'err')
  call to_history('n2',1,'err')
  call to_history('n3',1,'err')
  call to_history('d1',1.,'err')
  call to_history('o1',0.,'err')
  call to_history('esize',4,'err')

  nout=1
  select case(fdim)
  case (1)
     nout(1:ndim)=n(1:ndim)
     dout(1:ndim)=d(1:ndim)
     call to_hist_int_array('nout',nout,'outpar.P')
     call to_hist_float_array('dout',dout,'outpar.P')
  case (2)
     call to_history('nout2',n(1),'outpar.P')
     call to_history('dout2',d(1),'outpar.P')
     call to_history('nout3',n(2),'outpar.P')
     call to_history('dout3',d(2),'outpar.P')
  end select

  call sep_close()

  allocate(data1(nn),data2(nn))
  allocate(res(nn))

  ! filters same size everywhere
  allocate(nh(product(np)))
  nh(1:nnp)=product(nf)
  nhmax=maxval(nh)

  ! divide up into patches
  allocate (pch(nn))
  do i=1,nn
     pch(i)=0
     call line2cart(n,i,ii)
     do j=1,ndim
        ii(j)=(1+(ii(j)-1)/psize(j))
     end do
     call cart2line(np,ii,pch(i))
  end do

  ! allocate ns helix
  call nallocate(ff,nh,pch)
  allocate(ff%mis(nn))
  ff%mis=.false.
  do i=1,nnp
     select case(fdim)
     case(1)
        do j=1,size(ff%hlx(i)%flt)
           ff%hlx(i)%lag(j)=j-(nh(i)+1)/2
           ff%hlx(i)%flt(j)=0.
        end do
     case(2)
        do j=1,nf(2)
           do k=1,nf(1)
              index=(j-1)*nf(1)+k
              ff%hlx(i)%lag(index)=(k-(nf(1)+1)/2) + (j-(nf(2)+1)/2) * n(1)
              ff%hlx(i)%flt(index)=0.
           end do
        end do
     case default
        call seperr('dimensionality not implemented')
     end select
  end do
  write(0,*) ff%hlx(1)%lag
  nfilt=nhmax*nnp
  allocate(aa(nfilt),xmov(nfilt,niter))
  allocate(bb(nfilt))
  aa=0.
  bb=0.

  call sep_read(data1)
  call sep_read(data2,'master')

!!! APPLICATION SPECIFIC !!!

  ! Spatial roughener for preconditioning
  !  select case(1)
  select case(ndim)
  case (1)
     call allocatehelix(laplac,1)  
     laplac%lag(1)=nhmax;   laplac%flt(1)=-.99
  case (2)
     !     call allocatehelix(laplac,1)  
     !     laplac%lag(1)=np(1)*nhmax;   laplac%flt(1)=-.9

     call allocatehelix(laplac,3)  
     laplac%lag=(/ nhmax, (np(1)-1)*nhmax, nhmax*(np(1))  /)
     e1=0.2; e2=0.1
     laplac%flt=(/ e1-1.,      e2-1.,     (1.-e1)*(1.-e2)  /)
  case default
     call seperr('dimensionality not implemented')
  end select
  allocate(laplac%mis(nfilt))
  ff%mis=.false.

  ! find pef from data1
  call allocatehelix(pef1,5)
  do i=1,size(pef1%lag)
     pef1%lag(i)=i
  end do
  pef1%flt=0.
  allocate(pef1%mis(nn))
  pef1%mis=.false.
  call find_pef(data1,pef1,size(pef1%lag))

  ! whiten data1
  !  call helicon_init(pef1)
  !  stat=helicon_lop(.false.,.false.,data1,res)
  !  stat=helicon_lop(.true.,.false.,data1,res)

  call polydiv_init(nfilt,laplac)
  call ncnhconest_init(data1,ff,nhmax)
  call solver_prec(ncnhconest_lop,cgstep,polydiv_lop,nfilt,aa,data2,niter,eps,aa,res=res,verb=verb,err=err,xmov=xmov)

  ! write out filter coefficients

  ! whiten data1
  call polydiv_init(nfilt,pef1)
  stat=polydiv_lop(.false.,.false.,aa,bb)
  stat=polydiv_lop(.true.,.false.,aa,bb)

!  call sep_write(xmov,'filt')
  call sep_write(aa,'filt')
  call sep_write(err,'err')

  if (wantmatched) then
     ! load filter coefficients in to ff
     do i=1,nnp
        do j=1,size(ff%hlx(i)%flt)
           ff%hlx(i)%flt(j)=aa((i-1)*nhmax+j)
        end do
     end do

     ! apply match-filter to data
     call ncnhelicon_init(ff)
     stat=ncnhelicon_lop(.false.,.false.,data1,data2)
     call sep_write(data1)
  end if

end program MatchND


















