! Resample.x <in.H >out.H d1out= n1out= o1out=[o1in] factor=[8]
!
! Resample an arbitary axis to arbitrary n, d and o
!
! Band-limited interpolation
!	First over-samples by int(factor)
!	Then linear interpolates onto new grid
!	Works well for wavelets/seismic data etc
!
! Linear interpolation
!	Set factor=1
!	Works for velocity functions etc
!%

program Resample
  use sep 
  use fft
  use tohist

  use lint2
  use lint2skip
  implicit none

  real, pointer, dimension(:)   :: data,temp,coord1,prdata
  real, pointer, dimension(:,:) :: coord
  logical, pointer, dimension(:) :: thisaxis
  complex, pointer, dimension(:,:) :: cdata,pcdata
  integer :: outside,ndata,nn,ninterp,ntemp,inside,this,ndim,factor
  integer :: i,j,k,l,stat,m,thisout,thisin,farout,shift,tshif,pthis,p1

  integer, pointer, dimension(:) :: nin,nout,n,ii,ndef,nsave
  real,    pointer, dimension(:) :: din,oin, dout,oout,ddef,odef
  logical, parameter :: F=.false.
  
  call sep_init(SOURCE)
  ndim=sep_dimension()
  allocate (nin(ndim),nout(ndim),n(ndim),ii(ndim),din(ndim),oin(ndim))
  allocate (dout(ndim),oout(ndim),ndef(ndim),ddef(ndim),odef(ndim))
  allocate (thisaxis(ndim),nsave(ndim))
  call from_history ('n',nin)
write(0,*) 'N_IN = ',nin
  call from_history ('d',din)
  call from_history ('o',oin)
  call from_param ('factor',factor,1)
  factor=pad2(factor)
  nn=product(nin)

  ndef=-1
  ddef=-1
  odef=-1
  call from_param('nout',nout,ndef)
  call from_param('dout',dout,ddef)
  call from_param('oout',oout,odef)

write(0,*) 'N_OUT = ',nout

  ninterp=0
  thisaxis=.false.
  do i=1,ndim 
     if ((nout(i)>0).or.(dout(i)>0.).or.(oout(i)>0.)) then
        if ((nout(i)<0).and.(dout(i)<0.)) then
           nout(i)=nin(i)
           dout(i)=din(i)
write(0,*) 'case 1'
        else if (nout(i)<0) then
           nout(i)=(nin(i)*din(i))/dout(i) +1
write(0,*) 'case 2'
        else if (dout(i)<0.) then
           dout(i)=(nin(i)*din(i))/(nout(i))
write(0,*) 'case 3'
        end if
        ninterp=ninterp+1
        thisaxis(i)=.true.
        write(0,*) 'Interpolating axis ',i,' nin=',nin(i),' nout=',nout(i)
        write(0,*) '                  din=',din(i),' dout=',dout(i)
        write(0,*) '                  oin=',oin(i),' oout=',oout(i)
     end if
     if (nout(i)<0 ) nout(i)=nin(i)
     if (dout(i)<0.) dout(i)=din(i)
     if (oout(i)<0.) oout(i)=oin(i)
  end do

  allocate(ii(ninterp))
  j=0
  do i=1,ndim
     if (thisaxis(i)) then
        j=j+1
        ii(j)=i
     end if
  end do

write(0,*) 'N_OUT is ',nout
  call to_hist_int_array ('n',nout,'out')
  call to_hist_float_array ('d',dout,'out')
  call to_hist_float_array ('o',oout,'out')
  call sep_close()

  if (ninterp==0) then
     allocate(data(nin(1)))
     do i=1,product(nin)/nin(1)
        data=0.
        call sep_read(data)
        call sep_write(data)
     end do
     stop
  end if

  farout=product(nin)/product(nin(1:ii(ninterp)))
  write(0,*) 'farout=',farout,'     ninterp=',ninterp
  nsave=nin
  do i=1,farout
     ndata=nn/farout
     allocate(data(ndata))
     call sep_read(data)
     nin=nsave
     do j=1,ninterp

        thisin=nin(ii(j))
        thisout=nout(ii(j))
        inside=product(nin(1:ii(j)))/nin(ii(j))

        outside=ndata/(thisin*inside)
        ntemp=(ndata*thisout)/thisin
        allocate(temp(ntemp),coord(ntemp,2),coord1(ntemp))

        pthis=pad2(thisin)
        p1=pthis*factor

        allocate(cdata(inside,pthis),pcdata(inside,p1),prdata(inside*p1))

        do m=1,outside
           do k=1,thisout
              do l=1,inside
                 coord((m-1)*thisout*inside+(k-1)*inside+l,1)=real(l)
                 coord((m-1)*thisout*inside+(k-1)*inside+l,2)=(k-1)*dout(ii(j))+oout(ii(j))
              end do
           end do
        end do

        do m=1,outside
           shift=(m-1)*inside*thisin
           tshif=(m-1)*inside*thisout

           if (factor==1) then

              call lint2skip_init(inside,thisin,0.,1.,oin(ii(j)),din(ii(j))/factor,coord)
              stat=lint2skip_lop(F,F,data(1+shift:inside*thisin+shift),temp(1+tshif:inside*thisout+tshif))
           else

              do k=1,thisin
                 cdata(1:inside,k)=data(1+(k-1)*inside+shift:inside+(k-1)*inside+shift)
              end do
              cdata(:,thisin+1:pthis)=0.
              call ft2axis(.false.,.false.,cdata)
              pcdata=0.
              pcdata(:,(p1/2-pthis/2+1):(p1/2+pthis/2))=cdata(:,1:pthis)
              call ft2axis(.true.,.false.,pcdata)
              do k=1,p1
                 prdata(1+(k-1)*inside:inside+(k-1)*inside)=real(pcdata(:,k))
              end do

              call lint2skip_init(inside,p1,0.,1.,oin(ii(j)),din(ii(j))/factor,coord)
              stat=lint2skip_lop(F,F,prdata,temp(1+tshif:inside*thisout+tshif))
           end if
        end do

        deallocate(cdata,pcdata)
        deallocate(data)
        ndata=ntemp
        allocate(data(ndata))
        data=temp
        deallocate(temp,coord,coord1)
        nin(ii(j))=nout(ii(j))
     end do

     call sep_write(data)
  end do
end program Resample


